# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 解释器常量 ]

"""


KEYWORD_BASECODE_FILE_SUFFIX = 'basecode'
"""[ 关键字 :: 编码文件后缀 ]
@ 描述 : 底层编码文件的命名拼接字段。
"""

KEYWORD_IDENT_SCOPE_UNIVERSE = 'UNIVERSE'
"""[ 关键字 :: 公共作用域标识  ]
@ 描述 : 底层编码行的作用域标识, 表示作用于全域。
"""

SYMBOL_BASECODE_FILE_SEPARATOR = '-'
"""[ 符号 :: 编码文件间隔符  ]
@ 描述 : 底层编码文件的命名拼接间隔符号。
"""


# --------------------------------------------------------------------------
class StatementIdent:
    """ 
    [ 声明标识常量 ]

    ---
    描述
    - 底层编码行的声明事件参数标识。

    """
    Case = 'CASE'
    """ 声明用例标识 """

    Deco = 'DECO'
    """ 声明装饰器标识 """


# --------------------------------------------------------------------------
class GrammarSymbol:
    """ 
    [ 语法符号常量 ]

    ---
    描述
    - Awaken脚本语法中的语法符号。

    """
    Annotation = '##'
    """ 注释符 """

    Give = '='
    """ 赋值符 """

    Statement = '::'
    """ 声明符 """

    Run = '>>'
    """ 执行符 """

    Assert = '??'
    """ 断言符 """

    Call = '@'
    """ 方法调用符 """

    Quote = '$'
    """ 资源引用符 """

    VariablePath = '::'
    """ 字典路径应用符 """

    ScopePrefix = '?>'
    """ 作用域前缀 """


# --------------------------------------------------------------------------
class CodeLineType:
    """ 
    [ 编码行类型常量 ]

    """
    Give = 'GIVE'
    """ 赋值类型 """

    RGive = 'RGIVE'
    """ 执行赋值类型 """

    SCase = 'SCASE'
    """ 声明用例类型 """

    SDecorator = 'SDECORATOR'
    """ 声明装饰器类型 """

    Run = 'RUN'
    """ 执行类型 """

    Assert = 'ASSERT'
    """ 断言类型 """


# --------------------------------------------------------------------------
class CodeLineScopet:
    """ 
    [ 编码行作用域常量 ]

    """
    Global = 'GLOBAL'
    """ 全局 """

    Local = 'LOCAL'
    """ 局部 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ Awaken 命令行工具 ]

---
描述
    - 该模块用于通过命令行的方式调用 Awaken。

"""
import sys

from .baseic.common.template_rendering import G_TEMPLATE_RENDERING


# --------------------------------------------------------------------------
# 命令行工具 Main 入口
# --------------------------------------------------------------------------
def main() -> ...:
    command    = 'order'  # 现持有命令
    parameters = []       # 现持有参数


    # ----------------------------------------------------------------------
    if len(sys.argv) >= 2:
        command = sys.argv[1]
        argv    = sys.argv[2:]


    # ----------------------------------------------------------------------
    if command in ['order', '-order', '-o']:
        from .external.order import instruction_order
        instruction_order()


    # ----------------------------------------------------------------------
    elif command in ['init', '-init', '-i']:
        from .external.init_engineering import instruction_init_engineering
        instruction_init_engineering()


    # ----------------------------------------------------------------------
    elif command in ['make', '-make', '-m']:
        from .external.create_project import instruction_create_project
        instruction_create_project(argv)


    # ----------------------------------------------------------------------
    elif command in ['task', '-task', '-t']:
        from .external.create_task import instruction_create_task
        instruction_create_task(argv)


    # ----------------------------------------------------------------------
    elif command in ['run', '-run', '-r']:
        from .external.common import confirm_root_directory_runtime_project
        from .external.running_task import instruction_running_task
        confirm_root_directory_runtime_project()
        instruction_running_task(argv)


    # ----------------------------------------------------------------------
    elif command in ['server', '-server', '-s']:
        from .external.common import confirm_root_directory_runtime_project
        from .external.running_server import instruction_running_server
        confirm_root_directory_runtime_project()
        instruction_running_server(argv)


    # ----------------------------------------------------------------------
    elif command in ['help', '-help', '-h']:
        from .external.help import instruction_help
        instruction_help()


    # ----------------------------------------------------------------------
    elif command in ['version', '-version', '-v']:
        from .external.version import instruction_version
        instruction_version()


    # ----------------------------------------------------------------------
    elif command in ['update', '-update', '-u']:
        from .external.update import instruction_update
        instruction_update()


    # ----------------------------------------------------------------------
    else:
        G_TEMPLATE_RENDERING.render_print(
            title='Awaken',
            is_show_number=False,
            source=[
                f'没有这样的指令:: [{ command }] !'
            ],
        )
        exit(0)



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ 脚本全局方法 ]

"""
import yaml
import time
from pathlib import Path

from ..baseic.const import G_CONST
from ..baseic.broadcast.awaken_log import G_LOG
from ..baseic.decorator.baseic import deco_singleton


# --------------------------------------------------------------------------
@deco_singleton
class GlobalMethod:
    """
    [ 脚本全局方法 ]
    
    """


    def sleep(self, wait: int):
        """ 
        [ 等待 ]

        """
        time.sleep(wait)


    def logout(self, message: str, level=G_CONST.Common.LogLevel.Info) -> str:
        """ 
        [ 日志输出 ]

        ---
        描述
        - 输出日志到控制台。

        ---
        参数
        - message : { str } - 输出的日志信息。
        - level   : { str } - 日志等级。
        
        """
        G_LOG.out_map(level, message)
        return message


    def read_yaml(self, yaml_path: str) -> dict:
        """ 
        [ 读取YAML ]

        ---
        描述
        - 读取YAML文件数据。

        ---
        参数
        - yaml_path : { str } - YAML文件路径。
        
        ---
        返回
        - dict : YAML文件数据字典。

        """
        yaml_path = Path(yaml_path)
        
        with open(file=yaml_path.resolve(), mode='r', encoding='UTF-8') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        
        return data


# --------------------------------------------------------------------------
G_GLOBAL_METHOD: GlobalMethod = GlobalMethod()
""" 脚本全局方法全局实例 """



# ╔══════════════════════════════════════════════════════════════════════════╗
# ║ Copyright 2022. quinn.7@foxmail.com All rights reserved.                 ║
# ║                                                                          ║
# ║ Licensed under the Apache License, Version 2.0 (the "License");          ║
# ║ you may not use this file except in compliance with the License.         ║
# ║ You may obtain a copy of the License at                                  ║
# ║                                                                          ║
# ║ http://www.apache.org/licenses/LICENSE-2.0                               ║
# ║                                                                          ║
# ║ Unless required by applicable law or agreed to in writing, software      ║
# ║ distributed under the License is distributed on an "AS IS" BASIS,        ║
# ║ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. ║
# ║ See the License for the specific language governing permissions and      ║
# ║ limitations under the License.                                           ║
# ╚══════════════════════════════════════════════════════════════════════════╝
"""
[ Awaken Web Socket ]

"""
import asyncio
import websockets
import websockets_routes

# 初始化一个router对象
router = websockets_routes.Router()

tl = []
from random import randint
for _ in range(0, 100):
    tl.append(str(randint(1, 1000)))


@router.route("/awaken/{path}")
async def awaken_router(websocket, path):
    async for _ in websocket:
        if (path.params['path'] == 'test'):
            if len(tl) > 0:
                el = tl[0]
                del  tl[0]
                await websocket.send(f'信息:{el}')


async def main():
    async with websockets.serve(lambda x, y: router(x, y), "localhost", 8765):
        await  asyncio.Future()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 接口常量 ]

"""


# --------------------------------------------------------------------------
class StateCode:
    """ 
    [ 接口状态标识码常量 ]

    ---
    描述
    - 定义 WebApi 接口返回的接口错误状态标识。

    """
    Success = 0
    """ 请求成功 """

    Error   = -1
    """ 请求异常 """

    Warning = -2
    """ 服务器异常 """


# --------------------------------------------------------------------------
class StateMessage:
    """
    [ 接口状态提示语常量 ]

    """
    Success = 'success'
    """ 请求成功 """

    Error   = 'error'
    """ 请求异常 """

    Warning = 'warning'
    """ 服务器异常 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ SQLite语句处理器 ]

"""
from ...baseic.keyword import G_KEYWORD
from ...baseic.decorator.baseic import deco_singleton
from ...baseic.error.database import AwakenDatabaseTableNotFoundError
from ...baseic.error.database import AwakenDatabaseTableNotFoundFieldError
from ...baseic.error.database import AwakenDatabaseTableDeductionInserteError


class SqliteSelecter: ...
class SqliteInserter: ...
class SqliteUpdater : ...
class SqliteDeleter : ...


# --------------------------------------------------------------------------
@deco_singleton
class SqliteStatementHandle:
    """
    [ SQLite语句处理器 ]
    
    """
    _table_structure = {}
    """ 表结构映射 """


    # ----------------------------------------------------------------------
    def create_table(self, table_name: str, table_field: dict, if_not_exists: bool = True) -> str:
        """
        [ 建表 ]

        ---
        参数
        - table_name    : { str }  - 表名。
        - table_field   : { dict } - 表字段。
        - if_not_exists : { bool } - 建表时如果不存在则创建。

        ---
        返回
        - { str } : 建表语句。

        """
        # 表结构字典将保留一份表结构数据作为某些方法的数据格式处理依赖
        self._table_structure.update({table_name : {}})

        # 处理表字段语句
        table_field_sentence = ''
        for tfmn, tfmv in table_field.items():
            if len(table_field_sentence) > 0:
                table_field_sentence += ', \n'
            table_field_sentence += f'\t{tfmn} {tfmv}'
            self._table_structure[table_name].update({tfmn : tfmv}) # 更新表结构字典

        # 拼接返回
        not_exists_keyword = G_KEYWORD.Sqlite.CreateTable.IfNotExists if if_not_exists else ''
        return f'CREATE TABLE {not_exists_keyword} {table_name} \n(\n {table_field_sentence} \n);'


    # ----------------------------------------------------------------------
    def select(self, table_name: str) -> SqliteSelecter:
        """
        [ 表查询 ]

        ---
        参数
        - table_name : { str } - 表名。

        ---
        返回
        - { SqliteSelecter } : 指向表的SQLite选择器对象。

        """
        # 如果表不存在表结构中则抛出异常
        if table_name not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundError(table_name)

        return SqliteSelecter(table_name, self._table_structure[table_name])


    # ----------------------------------------------------------------------
    def insert(self, table_name: str) -> SqliteInserter:
        """
        [ 表插入 ]

        ---
        参数
        - table_name : { str } - 表名。

        ---
        返回
        - { SqliteInserter } : 指向表的SQLite插入操作器对象。

        """
        # 如果表不存在表结构中则抛出异常
        if table_name not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundError(table_name)

        return SqliteInserter(table_name, self._table_structure[table_name])


    # ----------------------------------------------------------------------
    def update(self, table_name: str) -> SqliteInserter:
        """
        [ 表更新 ]

        ---
        参数
        - table_name : { str } - 表名。

        ---
        返回
        - { SqliteUpdater } : 指向表的SQLite更新操作器对象。

        """
        # 如果表不存在表结构中则抛出异常
        if table_name not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundError(table_name)

        return SqliteUpdater(table_name, self._table_structure[table_name])


    # ----------------------------------------------------------------------
    def delete(self, table_name: str) -> SqliteDeleter:
        """
        [ 表删除 ]

        ---
        参数
        - table_name : { str } - 表名。

        ---
        返回
        - { SqliteDeleter } : 指向表的SQLite删除操作器对象。

        """
        # 如果表不存在表结构中则抛出异常
        if table_name not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundError(table_name)

        return SqliteDeleter(table_name, self._table_structure[table_name])


# --------------------------------------------------------------------------
class SqliteSelecter:
    """
    [ SQLite查询处理器 ]

    """
    _table_name: str
    """ 表名 """
    _table_structure: dict
    """ 表结构 """
    _query_items: dict
    """ 查询项 """
    _filter_items: dict
    """ 过滤项 """
    _sort_items: dict
    """ 排序项 """
    _page: int
    """ 页码 """
    _number: int
    """ 数据量 """
    _data_alias: list
    """ 数据别名 """


    # ----------------------------------------------------------------------
    def __init__(self, table_name: str, table_structure: dict) -> ...:
        """
        [ SQLite查询处理器 ]

        ---
        参数
        - table_name      : { str }  - 表名。
        - table_structure : { dict } - 表结构。

        """
        self._table_name = table_name
        self._table_structure = table_structure
        self._query_items = {}
        self._filter_items = {}
        self._sort_items = {}
        self._page   = None
        self._number = None
        self._data_alias = []


    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        statement_select = self._processing_query_statement()
        statement_filter = self._processing_filter_statement()
        statement_sort   = self._processing_sort_statement()
        statement_limit  = self._processing_data_limit_statement()
        return f'{statement_select} \nFROM \n\t{self._table_name} \n{statement_filter} \n{statement_sort} \n{statement_limit}'


    # ----------------------------------------------------------------------
    def get(self, field: str, event: str = None) -> ...:
        """
        [ 查询项 ]

        ---
        描述
        - 新增一个查询项, 执行语句后返回查询项数据。

        ---
        参数
        - field : { str } - 字段。
        - event : { str } - 查询事件, None表示不使用事件。
        
        """
        if field not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundFieldError(table_name=self._table_name, field_name=field)
        else:
            self._query_items.update({field : event})

    
    # ----------------------------------------------------------------------
    def filter_equal(self, field: str, value: str) -> ...:
        """
        [ 过滤相等 ]

        ---
        描述
        - 新增一个过滤项, 过滤掉字段与值相等的数据。

        ---
        参数
        - field : { str } - 字段。
        - value : { str } - 字段的值。

        """
        if field not in self._table_structure.keys():
            ...
        else:
            self._filter_items.update({field : value})


    # ----------------------------------------------------------------------
    def filter_equal_batch(self, field_dict: dict):
        for k, v in field_dict.items():
            self.filter_equal(k, v)


    # ----------------------------------------------------------------------
    def set_sort_asc(self, field: str):
        """
        [ 设置正序排序 ]

        """
        if field not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundFieldError(table_name=self._table_name, field_name=field)
        else:
            self._sort_items.update({field : G_KEYWORD.Sqlite.SortType.Asc})


    # ----------------------------------------------------------------------
    def set_sort_desc(self, field: str):
        """
        [ 设置逆序排序 ]

        """
        if field not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundFieldError(table_name=self._table_name, field_name=field)
        else:
            self._sort_items.update({field : G_KEYWORD.Sqlite.SortType.Desc})


    # ----------------------------------------------------------------------
    def set_data_limit(self, page, number):
        """
        [ 设置数据限制 ]

        """
        if page and number:
            self._page   = int(page)
            self._number = int(number)


    # ----------------------------------------------------------------------
    def _processing_query_statement(self):
        """
        [ 处理查询语句 ]

        """
        # 查询项为空则返回 SELECT *
        if len(self._query_items) == 0:
            for key in self._table_structure.keys():
                self._query_items.update({key: None})

        field_statement = ''
        for tfmn, tfmv in self._query_items.items():
            if len(field_statement) > 0:
                field_statement += ', '
            if tfmv == None:
                field_statement += tfmn
                self._data_alias.append(tfmn)
            else:
                field_statement += f'{tfmv}({tfmn})'
                self._data_alias.append(f'{tfmv.lower()}')

        return f'SELECT \n\t{field_statement}'


    # ----------------------------------------------------------------------
    def _processing_filter_statement(self):
        """
        [ 处理过滤语句 ]

        """
        # 过滤项为空则返回空字符串
        if len(self._filter_items) == 0:
            return ''

        field_statement = ''
        for wherek, wherev in self._filter_items.items():
            if len(field_statement) > 0:
                field_statement += '\nAND\n\t'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(wherek) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull
            ]:
                field_statement += f'{wherek} = {wherev}'
            else:
                field_statement += f"{wherek} = '{wherev}'"

        return f'WHERE \n\t{field_statement}'


    # ----------------------------------------------------------------------
    def _processing_sort_statement(self):
        """
        [ 处理排序语句 ]

        """
        if len(self._sort_items) == 0:
            return ''
        field_statement = ''
        for tfmn, tfmv in self._sort_items.items():
            if len(field_statement) > 0:
                field_statement += ', '
            field_statement += f'{tfmn} {tfmv}'
        return f'ORDER BY \n\t{field_statement}'


    # ----------------------------------------------------------------------
    def _processing_data_limit_statement(self):
        """
        [ 处理数据限制语句 ]

        """
        if self._page and self._number:
            # 计算偏移量
            limit_begin = 0
            if self._page > 1:
                limit_begin = (self._page - 1) * self._number
            return f'LIMIT \n\t{ limit_begin }, { self._number }'
        else:
            return ''


# --------------------------------------------------------------------------
class SqliteInserter:
    """
    [ SQLite插入处理器 ]

    """
    _with_lock: bool
    """ 是否带锁 """
    _table_name: str
    """ 表名 """
    _table_structure: dict
    """ 表结构 """
    _insert_items: dict
    """ 插入项 """


    # ----------------------------------------------------------------------
    def __init__(self, table_name: str, table_structure: dict) -> ...:
        """
        [ SQLite插入处理器 ]

        ---
        参数
        - table_name      : { str }  - 表名。
        - table_structure : { dict } - 表结构。

        """
        self._with_lock = False
        self._table_name = table_name
        self._table_structure = table_structure
        self._insert_items = {}


    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        statement_items  = self._processing_items_statement()
        statement_values = self._processing_values_statement()
        return f'INSERT INTO {self._table_name} \n(\n{statement_items}\n)\nVALUES \n(\n{statement_values}\n)'


    # ----------------------------------------------------------------------
    def lock(self) -> ...:
        """
        [ 带锁 ]

        """
        self._with_lock = True


    # ----------------------------------------------------------------------
    def deduction(self, values: list) -> ...:
        """
        [ 推导入表 ]

        ---
        描述
        - 自动根据表结构

        ---
        参数
        - values : { str } - 字段值列表。
        
        """
        table_structure_keys = self._table_structure.keys()
        if len(values) < len(table_structure_keys):
            raise AwakenDatabaseTableDeductionInserteError(
                self._table_name, len(table_structure_keys), len(values)
            )
        else:
            push_dict = dict(zip(table_structure_keys, values))
            for key, value in push_dict.items():
                self.push(key, value)


    # ----------------------------------------------------------------------
    def push(self, field: str, value: str = 'NULL') -> ...:
        """
        [ 插入项 ]

        ---
        描述
        - 新增一个插入项。

        ---
        参数
        - field : { str } - 字段。
        - value : { str } - 字段值。
        
        """
        if field not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundFieldError(table_name=self._table_name, field_name=field)
        else:
            self._insert_items.update({field : value})


    # ----------------------------------------------------------------------
    def _processing_items_statement(self):
        """
        [ 处理项语句 ]

        """
        # 过滤项为空则返回空字符串
        if len(self._insert_items) == 0:
            return ''

        statement = ''
        for itemn, _ in self._insert_items.items():
            if len(statement) > 0:
                statement += ', \n'
            statement += f'\t{itemn}'

        return statement


    # ----------------------------------------------------------------------
    def _processing_values_statement(self):
        """
        [ 处理值语句 ]

        """
        # 过滤项为空则返回空字符串
        if len(self._insert_items) == 0:
            return ''

        statement = ''
        for itemn, itemv in self._insert_items.items():
            if len(statement) > 0:
                statement += ', \n'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(itemn) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull,
            ]:
                statement += f'\t{itemv}'
            else:
                statement += f"\t'{itemv}'"

        return statement


# --------------------------------------------------------------------------
class SqliteUpdater:
    """
    [ SQLite更新处理器 ]

    """
    _with_lock: bool
    """ 是否带锁 """
    _table_name: str
    """ 表名 """
    _table_structure: dict
    """ 表结构 """
    _updata_items: dict
    """ 更新项 """
    _filter_items: dict
    """ 过滤项 """


    # ----------------------------------------------------------------------
    def __init__(self, table_name: str, table_structure: dict) -> ...:
        """
        [ SQLite更新处理器 ]

        ---
        参数
        - table_name      : { str }  - 表名。
        - table_structure : { dict } - 表结构。

        """
        self._with_lock = False
        self._table_name = table_name
        self._table_structure = table_structure
        self._updata_items = {}
        self._filter_items = {}


    # ----------------------------------------------------------------------
    def lock(self) -> ...:
        """
        [ 带锁 ]

        """
        self._with_lock = True


    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        statement_updatas = self._processing_updatas_statement()
        statement_filter  = self._processing_filter_statement()
        return f'UPDATE \n\t{self._table_name} \nSET\n{statement_updatas}\n{statement_filter}'


    # ----------------------------------------------------------------------
    def push(self, field: str, value: str = None) -> ...:
        """
        [ 更新项 ]

        ---
        描述
        - 新增一个更新项。

        ---
        参数
        - field : { str } - 字段。
        - value : { str } - 字段值。
        
        """
        if field not in self._table_structure.keys():
            raise AwakenDatabaseTableNotFoundFieldError(table_name=self._table_name, field_name=field)
        else:
            self._updata_items.update({field : value})


    # ----------------------------------------------------------------------
    def filter_equal(self, field: str, value: str) -> ...:
        """
        [ 过滤相等 ]

        ---
        描述
        - 新增一个过滤项, 过滤掉字段与值相等的数据。

        ---
        参数
        - field : { str } - 字段。
        - value : { str } - 字段的值。

        """
        if field not in self._table_structure.keys():
            ...
        else:
            self._filter_items.update({field : value})


    # ----------------------------------------------------------------------
    def filter_equal_batch(self, field_dict: dict):
        for k, v in field_dict.items():
            self.filter_equal(k, v)


    # ----------------------------------------------------------------------
    def _processing_updatas_statement(self):
        """
        [ 处理更新语句 ]

        """
        # 查询项为空则返回 SELECT *
        if len(self._updata_items) == 0:
            return ''

        field_statement = ''
        for tfmn, tfmv in self._updata_items.items():
            if len(field_statement) > 0:
                field_statement += ', \n'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(tfmn) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull
            ]:
                field_statement += f'\t{tfmn} = {tfmv}'
            else:
                field_statement += f"\t{tfmn} = '{tfmv}'"
                # tfmv = str(tfmv)
                # if tfmv[0] == '"' and tfmv[-1] == '"' or tfmv[0] == "'" and tfmv[-1] == "'":
                #     tfmv = ''.join(["'", tfmv[1: -1] ,"'"])
                #     field_statement += f'\t{tfmn} = {tfmv}'
                # else:
                #     field_statement += f"\t{tfmn} = '{tfmv}'"

        return f'{field_statement}'


    # ----------------------------------------------------------------------
    def _processing_filter_statement(self):
        """
        [ 处理过滤语句 ]

        """
        # 过滤项为空则返回空字符串
        if len(self._filter_items) == 0:
            return ''

        field_statement = ''
        for wherek, wherev in self._filter_items.items():
            if len(field_statement) > 0:
                field_statement += '\nAND\n\t'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(wherek) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull
            ]:
                field_statement += f'{wherek} = {wherev}'
            else:
                field_statement += f'{wherek} = "{wherev}"'

        return f'WHERE \n\t{field_statement}'


# --------------------------------------------------------------------------
class SqliteDeleter:
    """
    [ SQLite删除处理器 ]

    """
    _with_lock: bool
    """ 是否带锁 """
    _table_name: str
    """ 表名 """
    _table_structure: dict
    """ 表结构 """
    _filter_items: dict
    """ 过滤项 """


    # ----------------------------------------------------------------------
    def __init__(self, table_name: str, table_structure: dict) -> ...:
        """
        [ SQLite删除处理器 ]

        ---
        参数
        - table_name      : { str }  - 表名。
        - table_structure : { dict } - 表结构。

        """
        self._with_lock = False
        self._table_name = table_name
        self._table_structure = table_structure
        self._filter_items = {}


    # ----------------------------------------------------------------------
    def lock(self) -> ...:
        """
        [ 带锁 ]

        """
        self._with_lock = True


    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        statement_filter  = self._processing_filter_statement()
        return f'DELETE FROM \n\t{self._table_name} \n{statement_filter}'


    # ----------------------------------------------------------------------
    def filter_equal(self, field: str, value: str) -> ...:
        """
        [ 过滤相等 ]

        ---
        描述
        - 新增一个过滤项, 过滤掉字段与值相等的数据。

        ---
        参数
        - field : { str } - 字段。
        - value : { str } - 字段的值。

        """
        if field not in self._table_structure.keys():
            ...
        else:
            self._filter_items.update({field : value})


    # ----------------------------------------------------------------------
    def filter_equal_batch(self, field_dict: dict):
        for k, v in field_dict.items():
            self.filter_equal(k, v)


    # ----------------------------------------------------------------------
    def _processing_updatas_statement(self):
        """
        [ 处理更新语句 ]

        """
        # 查询项为空则返回 SELECT *
        if len(self._updata_items) == 0:
            return ''

        field_statement = ''
        for tfmn, tfmv in self._updata_items.items():
            if len(field_statement) > 0:
                field_statement += ', \n'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(tfmn) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull
            ]:
                field_statement += f'\t{tfmn} = {tfmv}'
            else:
                field_statement += f'\t{tfmn} = "{tfmv}"'

        return f'{field_statement}'


    # ----------------------------------------------------------------------
    def _processing_filter_statement(self):
        """
        [ 处理过滤语句 ]

        """
        # 过滤项为空则返回空字符串
        if len(self._filter_items) == 0:
            return ''

        field_statement = ''
        for wherek, wherev in self._filter_items.items():
            if len(field_statement) > 0:
                field_statement += '\nAND\n\t'

            # 如果过滤项是数据库数值类型则不加引号
            if self._table_structure.get(wherek) in [
                G_KEYWORD.Sqlite.DataType.IdKey,
                G_KEYWORD.Sqlite.DataType.Int,
                G_KEYWORD.Sqlite.DataType.IntNotNull
            ]:
                field_statement += f'{wherek} = {wherev}'
            else:
                field_statement += f'{wherek} = "{wherev}"'

        return f'WHERE \n\t{field_statement}'


_: ...

# --------------------------------------------------------------------------
G_SQLITE_STATEMENT_HANDLE = SqliteStatementHandle()
""" SQLite语句处理器全局实例 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
脚本全局方法映射

"""
import re

from .global_method import G_GLOBAL_METHOD


GLOBAL_METHOD_TYPE = type(G_GLOBAL_METHOD)
GLOBAL_METHOD_FUNCTION_MAP = {}


for fn, fv in GLOBAL_METHOD_TYPE.__dict__.items():
    if fn[0] != '_':
        fn = ''.join([n.title() for n in fn.rsplit('_')])
        cn_fn = re.findall(r'\[ (.*) \]', fv.__doc__)[0]
        GLOBAL_METHOD_FUNCTION_MAP.update({fn:fv})
        GLOBAL_METHOD_FUNCTION_MAP.update({cn_fn:fv})



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
命令行指令 - 更新版本

"""
import re
import subprocess

from .. import AwakenDetails
from ..baseic.common.template_rendering import G_TEMPLATE_RENDERING


# --------------------------------------------------------------------------
def instruction_update():
    """
    [ 命令行指令 - 更新版本 ]

    ---
    参数
    - argv : { list } - 参数列表。

    """
    G_TEMPLATE_RENDERING.render_print(
        title='Awaken - 更新版本',
        is_show_number=False,
        source=[
            '正在查询 PYPI 服务器 ...',
        ],
    )

    dos_result = subprocess.Popen(['pip', 'install', f'{ AwakenDetails.Name }=='], stderr=subprocess.PIPE)
    _, err = dos_result.communicate()
    newest_version: str = re.findall(r'from versions: (.*)\)', str(err))[0].split(',')[-1].strip()

    if newest_version > AwakenDetails.Version:
        G_TEMPLATE_RENDERING.render_print(
            title='Awaken - 更新版本',
            is_show_number=False,
            source=[
                f'从 PYPI 中发现最新版本号: {newest_version} !',
                '正在执行更新程序 ...',
                f'{AwakenDetails.Version} >> {newest_version}',
            ],
        )

        dos_result = subprocess.Popen(['pip', 'install', '--upgrade', AwakenDetails.Name], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        dos_result.communicate()

        G_TEMPLATE_RENDERING.render_print(
            title='Awaken - 更新版本',
            is_show_number=False,
            source=[
                f'当前为最新版本号:: {newest_version}',
                '已完成更新 ...',
            ],
        )

    else:
        G_TEMPLATE_RENDERING.render_print(
            title='Awaken - 更新版本',
            is_show_number=False,
            source=[
                f'当前为最新版本号:: {newest_version}',
                '暂无更新 ...',
            ],
        )
        exit(0)



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
命令行指令 - 通用模块

"""
from ..baseic.const import G_CONST
from ..baseic.common.template_rendering import G_TEMPLATE_RENDERING


_NUMBER_NESTING_LAYERS = 3
""" 工程目录嵌套最大层数 """


# --------------------------------------------------------------------------
def confirm_root_directory_runtime_project():
    """
    [ 确认运行时工程的根目录 ]

    """
    cwd = G_CONST.Path.ACTUALLY_CWD

    for _ in range(_NUMBER_NESTING_LAYERS):
        engineering_init = cwd.joinpath(G_CONST.Name.FileName.EngineeringInit)
        if engineering_init.exists():
            G_CONST.Path.ACTUALLY_CWD      = cwd
            G_CONST.Path.DirPath.Data      = cwd.joinpath(G_CONST.Name.DirName.EngineeringData)
            G_CONST.Path.DirPath.Logs      = G_CONST.Path.DirPath.Data.joinpath(G_CONST.Name.DirName.EngineeringLogs)
            G_CONST.Path.DirPath.BaseCode  = G_CONST.Path.DirPath.Data.joinpath(G_CONST.Name.DirName.EngineeringBaseCode)
            G_CONST.Path.FilePath.EngineeringInit     = cwd.joinpath(G_CONST.Name.FileName.EngineeringInit)
            G_CONST.Path.FilePath.EngineeringConfig   = G_CONST.Path.DirPath.Data.joinpath(G_CONST.Name.FileName.EngineeringConfig)
            G_CONST.Path.FilePath.EngineeringDatabase = G_CONST.Path.DirPath.Data.joinpath(G_CONST.Name.FileName.EngineeringDatabase)
            return True
        else:
            cwd = cwd.resolve().parent
            continue
    
    G_TEMPLATE_RENDERING.render_print(
        title='检查环境',
        is_show_number=False,
        source=[
            '当前运行路径不是Awaken工程结构路径;',
            '请在工程根目录中执行工程初始化命令:',
            'awaken -init'
        ]
    ) 
    exit(0)



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" 
数据库相关异常

"""
from .baseic import AwakenBaseError


# --------------------------------------------------------------------------
class AwakenDatabaseTableNotFoundError(AwakenBaseError):
    """ [ 数据库数据表不存在 ] """

    def __init__(
        self, 
        table_name: str,
    ):
        self.table_name = table_name
        super().__init__()

    def __str__(self) -> str:
        self.bas_message = [
            f'数据库中不存在数据表:: {self.table_name} !'
        ]
        return super().__str__()


# --------------------------------------------------------------------------
class AwakenDatabaseTableNotFoundFieldError(AwakenBaseError):
    """ [ 数据库数据表不存在字段 ] """

    def __init__(
        self, 
        table_name: str,
        field_name: str,
    ):
        self.table_name = table_name
        self.field_name = field_name
        super().__init__()

    def __str__(self) -> str:
        self.bas_message = [
            f'数据库表:: {self.table_name} 不存在字段 :: {self.field_name} !'
        ]
        return super().__str__()


# --------------------------------------------------------------------------
class AwakenDatabaseTableDeductionInserteError(AwakenBaseError):
    """ [ 数据库数据表推导插入失败 ] """

    def __init__(
        self,
        table_name: str,
        expect_para_count: int,
        actual_para_count: int,
    ):
        self.table_name = table_name
        self.expect_para_count = expect_para_count
        self.actual_para_count = actual_para_count
        super().__init__()

    def __str__(self) -> str:
        self.bas_message = [
            f'数据库表:: {self.table_name}',
            f'期望接受 value 数:: {self.expect_para_count}',
            f'实际接受 value 数:: {self.actual_para_count}'
        ]
        return super().__str__()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 加解密相关函数 ]

"""
import hashlib


# --------------------------------------------------------------------------
def encrypt_md5(string: str) -> str:
    """
    [ MD5加密 ]

    ---
    参数
    - string : { str } - 需要加密的字符串。

    ---
    返回
    - str : 加密后的字符串。

    """
    md5hash = hashlib.md5(string.encode('UTF-8'))
    return md5hash.hexdigest()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
WEB驱动引擎方法映射

"""
import re

from .web_engine import WebEngine


WEB_ENGINE_TYPE = type(WebEngine())
WEB_ENGINE_FUNCTION_MAP = {}

for fn, fv in WEB_ENGINE_TYPE.__dict__.items():
    if fn[0] != '_':
        fn = ''.join([n.title() for n in fn.rsplit('_')])
        cn_fn = re.findall(r'\[ (.*) \]', fv.__doc__)[0]
        WEB_ENGINE_FUNCTION_MAP.update({fn:fv})
        WEB_ENGINE_FUNCTION_MAP.update({cn_fn:fv})


# --------------------------------------------------------------------------
WEB_INSTRUCTIONS = [
    {
        'title': '浏览器',
        'list': [
            {
                'type': 'Instructions',
                'name': '页面跳转',
                'help': '跳转至指向的 URL 网站。',
                'drivingCharacter': 'Goto',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': 'URL',
                        'help': '需要访问的 URL 路径。',
                        'key': 'url',
                        'value': 'https://www.baidu.com/',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '页面刷新',
                'help': '浏览器刷新当前页面。',
                'drivingCharacter': 'Reload',
                'drivingParameter': [],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '页面回退',
                'help': '浏览器回退到上一页。',
                'drivingCharacter': 'Back',
                'drivingParameter': [],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '页面前进',
                'help': '浏览器前进到下一页。',
                'drivingCharacter': 'Forward',
                'drivingParameter': [],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '获取页面名称',
                'help': '获取当前页面中的 Title 属性。',
                'drivingCharacter': 'GetPageName',
                'drivingParameter': [],
                'whetherThereReturn': True,
                'namespaceScope': 'UNIVERSE',
                'namespaceKey': '',
            },
            {
                'type': 'Instructions',
                'name': '重置上下文',
                'help': '重启浏览器并清除上下文缓存。',
                'drivingCharacter': 'NewContext',
                'drivingParameter': [],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '设置浏览器尺寸',
                'help': '设置当前浏览器的窗口尺寸。',
                'drivingCharacter': 'SetWindowSize',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '宽度',
                        'help': '浏览器宽度。',
                        'key': 'width',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '高度',
                        'help': '浏览器高度。',
                        'key': 'height',
                        'value': '',
                    }
                ],
                'whetherThereReturn': False,
            },
        ],
    },
    {
        'title': '鼠标操作',
        'list': [
            {
                'type': 'Instructions',
                'name': '鼠标单击',
                'help': '模拟鼠标进行单击元素操作。',
                'drivingCharacter': 'Click',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '鼠标双击',
                'help': '模拟鼠标进行双击元素操作。',
                'drivingCharacter': 'DblClick',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '鼠标悬停',
                'help': '模拟鼠标悬停在元素上。',
                'drivingCharacter': 'Hover',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '悬停秒数',
                        'help': '悬停等待时长。',
                        'key': 'wait',
                        'value': '1',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '鼠标拖拽',
                'help': '模拟鼠标拖拽行为, 将元素1拖拽至元素2, 然后松开鼠标。',
                'drivingCharacter': 'DragTo',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素1选择器',
                        'help': '元素1选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element1',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素1下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid1',
                        'value': '1',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器2',
                        'help': '元素选择器2, 允许 selector 或 XPath 定位。',
                        'key': 'element2',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素2下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid2',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
        ],
    },
    {
        'title': '通用元素',
        'list': [
            {
                'type': 'Instructions',
                'name': '获取元素数量',
                'help': '获取当前页面中定位元素的数量。',
                'drivingCharacter': 'GetElementCount',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    }
                ],
                'whetherThereReturn': True,
                'namespaceScope': 'UNIVERSE',
                'namespaceKey': '',
            },
            {
                'type': 'Instructions',
                'name': '获取元素文本',
                'help': '获取当前页面中定位元素的文本。',
                'drivingCharacter': 'GetElementText',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': True,
                'namespaceScope': 'UNIVERSE',
                'namespaceKey': '',
            },
            {
                'type': 'Instructions',
                'name': '获取元素属性值',
                'help': '获取输入框控件中输入的值。',
                'drivingCharacter': 'GetAttribute',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '属性名',
                        'help': '需获取值的属性名称。',
                        'key': 'name',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': True,
                'namespaceScope': 'UNIVERSE',
                'namespaceKey': '',
            },
        ]
    },
    {
        'title': '输入框控件',
        'list': [
            {
                'type': 'Instructions',
                'name': '输入框输入',
                'help': '该函数期望的目标控件是 Input 类型',
                'drivingCharacter': 'Send',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '文本',
                        'help': '输入的值。',
                        'key': 'key',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '获取输入框的值',
                'help': '获取输入框控件中输入的值。',
                'drivingCharacter': 'GetInputValue',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': True,
                'namespaceScope': 'UNIVERSE',
                'namespaceKey': '',
            },
        ],
    },
    {
        'title': '弹窗控件',
        'list': [
            {
                'type': 'Instructions',
                'name': '弹窗确定',
                'help': '点击一个绑定了弹窗事件的按钮元素, 然后静默确认该弹窗。',
                'drivingCharacter': 'PopupOk',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    },
                    {
                        'type': 'Input',
                        'name': '填充文本',
                        'help': '如果该弹窗存在输入框, 可通过参数 text 来填充。',
                        'key': 'text',
                        'value': '',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '弹窗取消',
                'help': '点击一个绑定了弹窗事件的按钮元素, 然后静默取消该弹窗。',
                'drivingCharacter': 'PopupNo',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    },
                ],
                'whetherThereReturn': False,
            },
        ],
    },
    {
        'title': '其他',
        'list': [
            {
                'type': 'Instructions',
                'name': '上传文件',
                'help': '点击一个绑定了文件上传事件的按钮元素, 然后静默处理该事件。',
                'drivingCharacter': 'Upload',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '文件路径',
                        'help': '文件完整路径。',
                        'key': 'file_path',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '下载文件',
                'help': '点击一个绑定了文件下载事件的按钮元素, 然后静默处理该事件。',
                'drivingCharacter': 'Download',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '文件路径',
                        'help': '文件保存完整路径。',
                        'key': 'file_path',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素选择器',
                        'help': '元素选择器, 允许 selector 或 XPath 定位。',
                        'key': 'element',
                        'value': '',
                    },
                    {
                        'type': 'Input',
                        'name': '元素下标',
                        'help': '若定位到多个元素, 可通过元素下标指定其一。',
                        'key': 'eid',
                        'value': '1',
                    }
                ],
                'whetherThereReturn': False,
            },
            {
                'type': 'Instructions',
                'name': '保存图片',
                'help': '保存当前屏幕截图。',
                'drivingCharacter': 'SavePng',
                'drivingParameter': [
                    {
                        'type': 'Input',
                        'name': '图片保存路径',
                        'help': '屏幕截图后需要保存至的本机完整路径。',
                        'key': 'save_path',
                        'value': '',
                    }
                ],
                'whetherThereReturn': False,
            },
        ],
    },
]



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 数据库 - 任务代理 ]

"""
import time

from .. import DB
from .. import SQL
from ....baseic.structural.interpreter import AwakenTask
from ....baseic.error.server import AwakenDatabaseAgentError


# --------------------------------------------------------------------------
class DbTaskRecordBroker:
    """
    [ 数据库 - 任务代理 ]

    ---
    描述
    - 该类将代理任务执行过程中的数据库写入事务。
    
    """
    _task: AwakenTask
    """ 任务对象 """
    _task_record_id: int
    """ 任务记录ID """
    _case_record_id: int
    """ 用例记录ID """
    _task_created_date: str
    """ 任务创建日期 """
    _task_start_time: str
    """ 任务开始时间 """
    _task_start_timestamp: int
    """ 任务开始时间戳 """
    _case_start_time: str
    """ 用例开始时间 """
    _case_start_timestamp: int
    """ 用例开始时间戳 """


    # ----------------------------------------------------------------------
    def __init__(self, task: AwakenTask):
        self._task = task


    # ----------------------------------------------------------------------
    def created_task(self, name: str, docs: str) -> int:
        """
        [ 创建任务记录 ]

        ---
        描述
        - 写入数据库任务记录。

        ---
        参数
        - name : { str } - 任务名称。
        - docs : { str } - 任务注释。

        ---
        返回
        - int : 写入的任务记录ID。

        """
        # 记录任务开始时间
        self._task_created_date    = time.strftime('%Y-%m-%d')
        self._task_start_time      = time.strftime('%H:%M:%S')
        self._task_start_timestamp = int(time.time())

        # 录入数据库任务记录
        try:
            self._task_record_id = DB.execute(SQL.TaskRecord.creation(
                self._task.database_task_id,
                name,
                docs,
                self._task.task_type,
                self._task_created_date,
                self._task_start_time,
            ))
            
        except BaseException:
            raise AwakenDatabaseAgentError

        return self._task_record_id


    # ----------------------------------------------------------------------
    def created_case(self, name: str, docs: str) -> int:
        """
        [ 创建用例记录 ]

        ---
        描述
        - 写入数据库用例记录。

        ---
        参数
        - name : { str } - 用例名称。
        - docs : { str } - 用例注释。

        ---
        返回
        - int : 写入的用例记录ID。

        """
        # 记录用例开始时间
        self._case_start_time = time.strftime('%H:%M:%S')
        self._case_start_timestamp = int(time.time())

        # 录入数据库用例记录
        self._case_record_id = DB.execute(SQL.CaseRecord.creation(
            self._task_record_id,
            name,
            docs,
            self._task.task_type,
            self._task_created_date,
            self._case_start_time,
        ))

        return self._case_record_id


    # ----------------------------------------------------------------------
    def update_case(
        self, 
        result: str, 
        process: str
    ):
        """
        [ 更新用例记录 ]

        ---
        参数
        - result : { str } - 结果。
        - process : { str } - 过程。

        """
        # 更新数据库用例记录
        DB.execute(SQL.CaseRecord.update(
            self._case_record_id,
            result,
            time.strftime('%H:%M:%S'),
            (int(time.time()) - self._case_start_timestamp),
            process
        ))


    # ----------------------------------------------------------------------
    def update_task(self, result: str, process: str):
        """
        [ 更新任务记录 ]

        ---
        参数
        - result : { str } - 结果。
        - process : { str } - 过程。

        """
        # 更新数据库任务记录
        DB.execute(SQL.TaskRecord.update(
            self._task_record_id,
            result,
            time.strftime('%H:%M:%S'),
            (int(time.time()) - self._task_start_timestamp),
            process,
        ))

        # 保存事务
        DB.commit()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" 
[ WebApi装饰器 ]

"""
from ...server.api.server_request_handler import ServerRequestHandler


# --------------------------------------------------------------------------
def deco_webapi_error_capture(cls):
    """ 
    [ WebApi错误捕捉装饰器 ]

    ---
    描述
    - 冠以该装饰器的接口在请求异常时将返回错误请求响应。

    """
    def _deco_webapi_error_capture(*args, **kwargs):
        try:
            return cls(*args, **kwargs)
        except BaseException as error:
            return ServerRequestHandler.unsuccessful(str(error))
    return _deco_webapi_error_capture



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
WEB驱动引擎

"""
import time
import asyncio
from typing import Any

from pathlib import Path

from ...baseic.const import G_CONST
from ...baseic.error.engine import AwakenWebEngineError


# --------------------------------------------------------------------------
class WebEngine:
    """
    [ WEB驱动引擎 ]

    """
    _playwright: Any
    """ Playwright对象 """
    _browser_map: dict
    """ 浏览器类型映射字典 """
    _browser = None
    """ 浏览器进程对象 """
    _context = None
    """ 浏览器上下文对象 """
    _page = None
    """ 浏览器页面对象 """


    # ----------------------------------------------------------------------
    def _init_playwright(self, playwright) -> ...:
        self._playwright = playwright
        self._browser_map = {
            G_CONST.Type.Browser.Webkit : self._playwright.webkit,
            G_CONST.Type.Browser.Firefox : self._playwright.firefox,
            G_CONST.Type.Browser.Chromium : self._playwright.chromium,
        }


    # ----------------------------------------------------------------------
    async def _browser_start(
        self, 
        browser_type=G_CONST.Type.Browser.Chromium, 
        headless=False
    ) -> ...:
        """ 
        [ 启动浏览器 ]

        ---
        描述
        - 根据传入的浏览器类型启动对象浏览器进程。

        ---
        参数
        - browser_type : { str } - 浏览器类型, 默认谷歌浏览器。
        - headless : { bool } - 是否以无头模式运行。

        """
        # 如果浏览器类型不存在, 则使用默认的 Chromium 浏览器
        if browser_type not in self._browser_map.keys():
            browser_type = G_CONST.Type.Browser.Chromium

        # 创建浏览器实例不存在则创建
        if not self._browser:
            self._browser = await self._browser_map[browser_type].launch(headless=headless)
            self._context = await self._browser.new_context()


    # ----------------------------------------------------------------------
    async def _page_start(self) -> ...:
        """ 
        [ 启动页面 ]

        ---
        描述
        - 浏览器进程启动页面实例。

        """
        self._page = await self._context.new_page()


    # ----------------------------------------------------------------------
    async def _page_close(self) -> ...:
        """
        [ 关闭页面 ]

        ---
        描述
        - 浏览器进程关闭页面实例。

        """
        if self._page:
            await self._page.close()
            self._page = None


    # ----------------------------------------------------------------------
    async def _exit(self) -> ...:
        """ 
        [ 退出浏览器 ]

        ---
        描述
        - 关闭页面、上下文、浏览器对象。
        
        """
        if self._page:
            await self._page.close()
            await self._context.close()
            await self._browser.close()
        self._page = None
        self._context = None
        self._browser = None


    # ----------------------------------------------------------------------
    async def new_context(self) -> ...:
        """ 
        [ 重置上下文 ]

        ---
        描述
        - 重置浏览器的上下文, 然后重新打开一个新页面。
        
        """
        await self._page_close()
        await self._context.close()
        self._context = await self._browser.new_context()
        self._page = await self._context.new_page()


    # ----------------------------------------------------------------------
    async def get_page_name(self) -> str:
        """ 
        [ 获取页面名称 ]

        ---
        描述
        - 获取当前页面中的 Title 属性。

        ---
        返回
        - str : 当前页面的名称。

        """
        return await self._page.title()


    # ----------------------------------------------------------------------
    async def get_element_count(self, element: str) -> int:
        """
        [ 获取元素数量 ]

        ---
        描述
        - 获取当前页面中定位元素的数量。

        ---
        参数
        - element { str } : 元素选择器。

        ---
        返回
        - int : 元素数量。

        """
        count = await self._locator_element_count(element)
        return count


    # ----------------------------------------------------------------------
    async def get_element_text(self, element: str, eid: int = 1) -> str:
        """ 
        [ 获取元素文本 ]

        ---
        描述
        - 获取当前页面中定位元素的文本。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { str } : 元素下标。

        ---
        返回
        - str : 元素的文本。

        """
        element = await self._locator_element(element, eid)
        element_text = await element.text_content()
        return element_text

    
    # ----------------------------------------------------------------------
    async def get_input_value(self, element: str, eid: int = 1) -> str:
        """
        [ 获取输入框的值 ]

        ---
        描述
        - 获取输入框控件中输入的值。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { str } : 元素下标。

        ---
        返回
        - str : 输入框的文本。

        """
        element = await self._locator_element(element, eid)
        element_text = await element.input_value()
        return element_text


    # ----------------------------------------------------------------------
    async def get_attribute(self, name: str, element: str, eid: int = 1) -> str | int:
        """ 
        [ 获取元素属性值 ]

        ---
        描述
        - 获取元素指定属性的值。

        ---
        参数
        - name { str } : 需获取值的属性名称。
        - element { str } : 元素选择器。
        - eid { str } : 元素下标。

        ---
        返回
        - str : 元素的属性值。

        """
        element = await self._locator_element(element, eid)
        element_value = await element.get_attribute(name)
        return element_value


    # ----------------------------------------------------------------------
    async def set_window_size(self, width: int, height: int) -> ...:
        """ 
        [ 设置浏览器尺寸 ]

        ---
        描述
        - 设置当前浏览器的窗口尺寸。

        ---
        参数
        - width { int } : 浏览器宽度。
        - height { int } : 浏览器高度。

        """
        size = \
        {
            'width' : int(width),
            'height': int(height),
        }
        await self._page.set_viewport_size(size)


    # ----------------------------------------------------------------------
    async def goto(self, url: str):
        """ 
        [ 页面跳转 ]

        ---
        描述
        - 访问输入的 URL 网站。

        ---
        参数
        - url { str } : 需要访问的 URL 路径。

        """
        await self._page.goto(url)


    # ----------------------------------------------------------------------
    async def back(self):
        """ 
        [ 页面回退 ]

        ---
        描述
        - 浏览器回退到上一页。

        """
        await self._page.go_back()


    # ----------------------------------------------------------------------
    async def forward(self):
        """ 
        [ 页面前进 ]

        ---
        描述
        - 浏览器前进到下一页。

        """
        await self._page.go_forward()

    
    # ----------------------------------------------------------------------
    async def reload(self):
        """ 
        [ 页面刷新 ]

        ---
        描述
        - 浏览器刷新当前页面。
            
        """
        await self._page.reload()


    # ----------------------------------------------------------------------
    async def send(self, key: str, element: str, eid: int = 1):
        """ 
        [ 输入框输入 ]

        ---
        描述
        - 该函数期望的目标控件是 Input 类型。

        ---
        参数
        - key { str } : 输入的值。
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        element = await self._locator_element(element, eid)
        if not isinstance(key, str):
            key = str(key)
        await element.fill(key)


    # ----------------------------------------------------------------------
    async def click(self, element: str, eid: int = 1):
        """ 
        [ 鼠标单击 ]

        ---
        描述
        - 模拟鼠标进行单击元素操作。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        element = await self._locator_element(element, eid)
        await element.click()


    # ----------------------------------------------------------------------
    async def dbl_click(self, element, eid: int = 1):
        """ 
        [ 鼠标双击 ]

        ---
        描述
        - 模拟鼠标进行双击元素操作。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        element = await self._locator_element(element, eid)
        await element.dblclick()


    # ----------------------------------------------------------------------
    async def drag_to(self, element1: str, element2: str, eid1: int = 1, eid2: int = 1):
        """ 
        [ 鼠标拖拽 ]

        ---
        描述
        - 模拟鼠标拖拽行为, 将元素1拖拽至元素2, 然后松开鼠标。

        ---
        参数
        - element1 { str } : 元素1选择器。
        - element2 { str } : 元素2选择器。
        - eid1 { int } : 元素1下标。
        - eid2 { int } : 元素2下标。

        """
        element1 = await self._locator_element(element1, eid1)
        element2 = await self._locator_element(element2, eid2)
        await element1.drag_to(element2)


    # ----------------------------------------------------------------------
    async def hover(self, wait: int, element: str, eid: int = 1):
        """ 
        [ 鼠标悬停 ]

        ---
        描述
        - 模拟鼠标悬停在元素上。

        ---
        参数
        - wait { int } : 悬停等待时长。
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        element = await self._locator_element(element, eid)
        await element.hover()
        await asyncio.sleep(wait)


    # ----------------------------------------------------------------------
    async def popup_ok(self, element: str, eid: int = 1, text: str = ''):
        """ 
        [ 弹窗确定 ]

        ---
        描述
        - 点击一个绑定了弹窗事件的按钮元素, 然后静默确认该弹窗, 如果该弹窗存在输入框, 可通过参数 text 来填充。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。
        - text { str } : 如果该弹窗存在输入框, 可通过参数 text 来填充。

        """
        # 弹窗处理函数
        async def event(dialog):
            global temporary_text_register
            temporary_text_register = dialog.message
            await dialog.accept(text)

        # 绑定弹窗处理事件
        self._page.once('dialog', event)
        element = await self._locator_element(element, eid)
        await element.click()
        return temporary_text_register


    # ----------------------------------------------------------------------
    async def popup_no(self, element: str, eid: int = 1):
        """ 
        [ 弹窗取消 ]

        ---
        描述
        - 点击一个绑定了弹窗事件的按钮元素, 然后静默取消该弹窗。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        # 弹窗处理函数
        async def event(dialog):
            global temporary_text_register
            temporary_text_register = dialog.message
            await dialog.dismiss()

        # 绑定弹窗处理事件
        self._page.once('dialog', event)
        element = await self._locator_element(element, eid)
        await element.click()
        return temporary_text_register


    # ----------------------------------------------------------------------
    async def upload(self, file_path: str, element: str, eid: int = 1):
        """ 
        [ 上传文件 ]

        ---
        描述
        - 点击一个绑定了文件上传事件的按钮元素, 然后静默处理该事件。

        ---
        参数
        - file_path { str } : 文件完整路径。
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        # 监听文件选择器对象
        self._page.once('filechooser', lambda _: ...)
        async with self._page.expect_file_chooser() as chooser_object:
            element = await self._locator_element(element, eid)
            await element.click()

        chooser_object = await chooser_object.value
        await chooser_object.set_files(file_path)


    # ----------------------------------------------------------------------
    async def download(self, file_path: str, element: str, eid: int = 1):
        """ 
        [ 下载文件 ]

        ---
        描述
        - 点击一个绑定了文件下载事件的按钮元素, 然后静默处理该事件。

        ---
        参数
        - file_path { str } : 文件保存完整路径。
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        """
        # 监听下载对象
        self._page.once('download', lambda _: ...)
        async with self._page.expect_download() as download_object:
            element, _ = await self._locator_element(element, eid)
            await element.click()

        download_object = await download_object.value
        await download_object.save_as(file_path)

    
    # ----------------------------------------------------------------------
    async def save_png(self, save_path: str):
        """ 
        [ 保存图片 ]

        ---
        描述
        - 保存当前屏幕截图。

        ---
        参数
        - save_path { str } : 屏幕截图后需要保存至的本机完整路径。

        """
        png_bytes = await self._page.screenshot()
        suffix = '.png'
        if save_path[-4:] != suffix:
            save_path = ''.join([save_path, suffix])

        save_path = Path(save_path)
        save_path.touch(0o777)
        save_path.write_bytes(png_bytes)

        return save_path.resolve()


    # ----------------------------------------------------------------------
    async def _dispatch_event(self, element, event_name):
        await self._page.dispatch_event(element, event_name)


    # ----------------------------------------------------------------------
    async def _java_script(self, expression, args):
        return await self._page.evaluate(expression, args)


    # ----------------------------------------------------------------------
    async def _locator_element(self, element: str, eid: int) -> Any:
        """ 
        [ 定位元素 ]

        ---
        描述
        - 定位当前页面的所有符合预期的元素对象。

        ---
        参数
        - element { str } : 元素选择器。
        - eid { int } : 元素下标。

        ---
        返回
        - 元素对象。

        """
        element_count  = 0
        retries_number = 5  # 元素重试次数
        retries_sleep  = 1  # 元素重试间隔时间

        for _ in range(retries_number):
            result = self._page.locator(element)
            element_count = await result.count()
            if element_count < 1:
                time.sleep(retries_sleep)
                continue
            else:
                break
        
        if element_count == 0:
            raise AwakenWebEngineError(f'无法定位到元素:: { element } !')
        if element_count < eid:
            raise AwakenWebEngineError(f'元素列表数量::{ element_count }, 但是传递ID::{ eid } !')

        return result.nth(eid-1)

        # try:
        #     for _ in range(retries_number):
        #         result = self._page.locator(element)
        #         element_count = await result.count()
        #         if element_count < 1:
        #             time.sleep(retries_sleep)
        #             continue
        #         else:
        #             break
                
        #     if element_count == 0:
        #         raise BaseException

        # except BaseException:
        #     raise AwakenWebEngineError(f'无法定位到元素 :: { element } !')
                
        # try:
        #     # 如果 eid 大于元素数量则抛出异常
        #     if element_count < eid: 
        #         raise BaseException
        #     else:
        #         return result.nth(eid - 1), element_count

        # except BaseException:
        #     raise AwakenWebEngineError(f'EID 超出元素对象数量, 无法应用 !')


    # ----------------------------------------------------------------------
    async def _locator_element_count(self, element: str) -> int:
        """ 
        [ 定位元素数量 ]

        ---
        参数
        - element { str } : 元素选择器。
        
        ---
        返回
        - 元素数量。
        
        """
        result = self._page.locator(element)
        element_count = await result.count()
        return element_count



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 常量索引 ]

"""
from . import const_api
from . import const_common
from . import const_interpreter
from . import const_name
from . import const_path
from . import const_state
from . import const_type


# --------------------------------------------------------------------------
class AwakenConst:
    """ 
    [ 常量索引 ]

    """


    # ----------------------------------------------------------------------
    @property
    def Api(self) -> const_api:
        """ 
        [ 接口常量 ]

        """
        return const_api


    # ----------------------------------------------------------------------
    @property
    def Common(self) -> const_common:
        """ 
        [ 通用常量 ]

        """
        return const_common


    # ----------------------------------------------------------------------
    @property
    def Interpreter(self) -> const_interpreter:
        """ 
        [ 解释器常量 ]

        """
        return const_interpreter


    # ----------------------------------------------------------------------
    @property
    def Name(self) -> const_name:
        """ 
        [ 名称常量 ]

        """
        return const_name


    # ----------------------------------------------------------------------
    @property
    def Path(self) -> const_path:
        """ 
        [ 路径常量 ]

        """
        return const_path


    # ----------------------------------------------------------------------
    @property
    def State(self) -> const_state:
        """ 
        [ 状态常量 ]

        """
        return const_state


    # ----------------------------------------------------------------------
    @property
    def Type(self) -> const_type:
        """ 
        [ 类型常量 ]

        """
        return const_type


# --------------------------------------------------------------------------
G_CONST: AwakenConst  = AwakenConst()
""" 常量索引全局实例 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 脚本执行器相关接口 ]

"""
import json

from flask import Response
from flask import Blueprint

from ..server_request_handler import URL_PREFIX
from ..server_request_handler import ServerRequestHandler
from ...db import DB
from ...db import SQL
from ....baseic.const import G_CONST
from ....baseic.keyword import G_KEYWORD
from ....baseic.common.configure import G_CONFIGURE
from ....baseic.structural.interpreter import AwakenTask
from ....baseic.decorator.web_api import deco_webapi_error_capture
from ....core.perform.pool_runtime import POOL_RUNTIME
from ....core.interpreter.grammar_parser import GRAMMAR_PARSER
from ....kit.engine_agency import EngineAgency


# 创建驱动代理
for _ in range(G_CONFIGURE.get(G_KEYWORD.Common.Config.EngineQueueMaxCount)):
    POOL_RUNTIME.engine_queue.put(EngineAgency())
    POOL_RUNTIME.sign_engine()


blueprint_runer = Blueprint('runer', __name__)
""" [ 脚本执行器接口蓝图实例 ] """


POST_RUNER_RUNNING_TASK    = '/runer/runing/task'
""" 接口URL - 运行任务 """
POST_RUNER_RUNNING_TASKS   = '/runer/runing/tasks'
""" 接口URL - 批量运行任务 """
POST_RUNER_RUNNING_PROJECT = '/runer/runing/project'
""" 接口URL - 运行项目 """


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_runer.route(''.join([URL_PREFIX, POST_RUNER_RUNNING_TASK]), methods=['POST'])
def post_running_task() -> Response:
    """
    [ POST - 运行任务 ]

    ---
    必要参数
    - tid : { int } - 任务编号。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid
        ]
    )
    tid = paras.get(G_KEYWORD.Api.TasksFieldKey.Tid)

    # 从数据库中获取数据
    db_data = DB.execute(SQL.Tasks.get_task_script(tid))
    task_info = db_data.get(G_KEYWORD.Common.TransactionFieldKey.Items)[0]
    task_type = task_info.get(G_KEYWORD.DataBase.TasksFieldKey.Type)
    task_name = task_info.get(G_KEYWORD.DataBase.TasksFieldKey.Name)
    task_docs = task_info.get(G_KEYWORD.DataBase.TasksFieldKey.Docs)
    script_json = json.loads(task_info.get(G_KEYWORD.DataBase.TasksFieldKey.ScriptJson))

    # 解析成 Awaken 语法
    awaken_script_code = []
    awaken_script_code.append(f'TaskType = \'{task_type}\'')
    awaken_script_code.append(f'TaskName = \'{task_name}\'')
    awaken_script_code.append(f'TaskDocs = \'{task_docs}\'')

    if len(script_json['cases']) > 0:
        for case in script_json['cases']:
            # 声明用例
            case_statement = f'CASE :: {case["name"]}'
            if case['docs'] != '':
                case_statement += f' :: {case["docs"]}'
            awaken_script_code.append(case_statement)

            if len(case['logic']) > 0:
                for logic in case['logic']:
                    # TODO 暂未处理不同的逻辑块类型
                    driving_character = logic['list'][0]['drivingCharacter']
                    driving_parameter = []
                    for para_value in logic['list'][0]['drivingParameter']:
                        if para_value['value'] != '':
                            driving_parameter.append(para_value['value'])
                    
                    run_instructions = '    '
                    if logic['list'][0]['whetherThereReturn']:
                        if logic['list'][0]['namespaceKey'] != '':
                            if logic['list'][0]['namespaceScope'] == 'UNIVERSE':
                                run_instructions += '$'
                            run_instructions += f"{logic['list'][0]['namespaceKey']} = "
                        
                    run_instructions += f'{ driving_character }'
                    for para in driving_parameter:
                        run_instructions += f'  >>  {para}'

                    awaken_script_code.append(run_instructions)
            awaken_script_code.append('    ?? True')

    print(awaken_script_code)
    task_object = GRAMMAR_PARSER.parsing_json(awaken_script_code, task_type, tid)
    POOL_RUNTIME.task_queue.put(task_object)

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Message: '任务推送完毕 !',
    })
    return ServerRequestHandler.successful(result)


# @blueprint_runer.route(''.join([URL_PREFIX, TASK_URL_RUNNING_TASK_LIST]), methods=['POST'])
# def running_task_list():
#     """
#     [ 执行任务列表 ]

#     ---
#     必要参数:
#         tid { int } : 任务ID

#     """
#     try:
#         result = {}
#         # 解析请求参数
#         paras = ServerRequestHandler.analysis_request_parameter(
#             keys=['tid_list'],
#             must_keys=['tid_list']
#         )

#         for tid in paras['tid_list']:

#             # 从数据库中获取数据
#             task_script = DB.lock_read(SQL.Tasks.get_task_script(tid))[0][7]
#             script_json = json.loads(task_script)

#             # 解析成 Awaken 语法
#             awaken_script_code = []
#             awaken_script_code.append(f'TaskType = \'{script_json["type"]}\'')
#             awaken_script_code.append(f'TaskName = \'{script_json["name"]}\'')
#             awaken_script_code.append(f'TaskDocs = \'{script_json["docs"]}\'')

#             if len(script_json['cases']) > 0:
#                 for case in script_json['cases']:
#                     # 声明用例
#                     case_statement = f'CASE :: {case["name"]}'
#                     if case['docs'] != '':
#                         case_statement += f' :: {case["docs"]}'
#                     awaken_script_code.append(case_statement)

#                     if len(case['logic']) > 0:
#                         for logic in case['logic']:
#                             # TODO 暂未处理不同的逻辑块类型
#                             driving_character = logic['list'][0]['drivingCharacter']
#                             driving_parameter = []
#                             for para_value in logic['list'][0]['drivingParameter']:
#                                 if para_value['value'] != '':
#                                     driving_parameter.append(para_value['value'])
                            
#                             run_instructions = '    '
#                             if logic['list'][0]['whetherThereReturn']:
#                                 if logic['list'][0]['namespaceKey'] != '':
#                                     if logic['list'][0]['namespaceScope'] == 'UNIVERSE':
#                                         run_instructions += '$'
#                                     run_instructions += f"{logic['list'][0]['namespaceKey']} = "
                                
#                             run_instructions += f'{ driving_character }'
#                             for para in driving_parameter:
#                                 run_instructions += f'  >>  {para}'

#                             awaken_script_code.append(run_instructions)
#                     awaken_script_code.append('    ?? True')

#             task_object = GRAMMAR_PARSER.parsing_json(awaken_script_code, script_json['type'])
#             POOL_RUNTIME.task_queue.put(task_object)

#         # 返回值
#         result.update({'result': 'ok'})
#         return ServerRequestHandler.successful(result)

#     except BaseException as error:
#         return ServerRequestHandler.unsuccessful(str(error))



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 编辑器相关接口 ]

"""
import time
import json

from flask import Response
from flask import Blueprint

from ..server_request_handler import URL_PREFIX
from ..server_request_handler import ServerRequestHandler
from ...db import DB
from ...db import SQL
from ....baseic.const import G_CONST
from ....baseic.keyword import G_KEYWORD
from ....baseic.decorator.web_api import deco_webapi_error_capture


blueprint_editor = Blueprint('editor', __name__)
""" 编辑器接口蓝图 """


POST_EDITOR_TASKS_SCRIPT = '/editor/task_script'
""" 接口URL - 获取任务脚本 """
POST_EDITOR_SAVE_SCRIPT = '/editor/save_script'
""" 接口URL - 获取任务脚本 """ 


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_editor.route(''.join([URL_PREFIX, POST_EDITOR_TASKS_SCRIPT]), methods=['POST'])
def post_tasks_script() -> Response:
    """
    [ POST - 获取任务脚本 ]

    ---
    必要参数
    - tid : { str } - 任务编号。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid,
        ]
    )
    tid = paras.get(G_KEYWORD.Api.TasksFieldKey.Tid)

    # 从数据库中获取数据
    db_data = DB.execute(SQL.Tasks.get_task_script(tid=tid))
    item = db_data.get(G_KEYWORD.Common.TransactionFieldKey.Items)[0]
    script_json = item.get(G_KEYWORD.DataBase.TasksFieldKey.ScriptJson).replace("'", '"')

    # 处理返回
    result.update({
        G_KEYWORD.Api.TasksFieldKey.Type: item.get(G_KEYWORD.DataBase.TasksFieldKey.Type),
        G_KEYWORD.Api.TasksFieldKey.Name: item.get(G_KEYWORD.DataBase.TasksFieldKey.Name),
        G_KEYWORD.Api.TasksFieldKey.Docs: item.get(G_KEYWORD.DataBase.TasksFieldKey.Docs),
        G_KEYWORD.Api.TasksFieldKey.ScriptJson: script_json
    })
    return ServerRequestHandler.successful(result)


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_editor.route(''.join([URL_PREFIX, POST_EDITOR_SAVE_SCRIPT]), methods=['POST'])
def post_save_script() -> Response:
    """
    [ POST - 保存任务脚本 ]

    ---
    必要参数
    - tid : { str } - 任务编号。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid,
            G_KEYWORD.Api.TasksFieldKey.Name,
            G_KEYWORD.Api.TasksFieldKey.Docs,
            G_KEYWORD.Api.TasksFieldKey.ScriptJson,
        ]
    )
    script_json = paras.get(G_KEYWORD.Api.TasksFieldKey.ScriptJson)

    # 写入数据库
    DB.execute(SQL.Tasks.modify(
        tid=paras.get(G_KEYWORD.Api.TasksFieldKey.Tid),
        name=paras.get(G_KEYWORD.Api.TasksFieldKey.Name),
        docs=paras.get(G_KEYWORD.Api.TasksFieldKey.Docs),
        script_json_string=script_json
    ))
    DB.commit()

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Message: '任务保存成功 !'
    })
    return ServerRequestHandler.successful(result)



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ Awaken 数据库 ]

"""
import sqlite3
from typing import Any
from threading import Lock

from .sqlite_statement_handle import SqliteUpdater
from .sqlite_statement_handle import SqliteDeleter
from .sqlite_statement_handle import SqliteSelecter
from .sqlite_statement_handle import SqliteInserter
from .sqlite_statement_handle import G_SQLITE_STATEMENT_HANDLE
from ...baseic.const import G_CONST
from ...baseic.keyword import G_KEYWORD
from ...baseic.error.server import AwakenDataBaseError
from ...baseic.decorator.baseic import deco_singleton


# --------------------------------------------------------------------------
@deco_singleton
class AwakenDb(object):
    """
    [ Awaken 数据库 ]

    ---
    描述
    - 基于 SQLite 数据库封装了一些常用功能。

    """
    _conn: sqlite3.Connection = None
    """ 数据库连接对象 """
    _lock: Lock = Lock()
    """ 数据库互斥锁 """


    # ----------------------------------------------------------------------
    def __init__(self):
        try:
            # 创建/连接数据库
            if not self._conn:
                self.connect()
                self.commit()
 
        except Exception as error:
            raise AwakenDataBaseError(str(error))


    # ----------------------------------------------------------------------
    def execute(self, sql_handle: Any, with_lock: bool = False, return_data_type: str = 'DICT'):
        """
        [ 执行SQLitem句柄 ]

        """
        # print(str(sql_handle))

        if isinstance(sql_handle, SqliteSelecter):
            return self._execute_sqlite_handle_query(sql_handle, with_lock, return_data_type)
        
        elif isinstance(sql_handle, SqliteInserter):
            return self._execute_sqlite_handle_write(sql_handle, with_lock)

        elif isinstance(sql_handle, SqliteUpdater):
            return self._execute_sqlite_handle_write(sql_handle, with_lock)

        elif isinstance(sql_handle, SqliteDeleter):
            return self._execute_sqlite_handle_write(sql_handle, with_lock)
            
        else:
            raise AwakenDataBaseError('无效的SQLitem句柄对象 !')


    # ----------------------------------------------------------------------
    def _execute_sqlite_handle_query(self, sql_handle: Any, with_lock: bool = False, return_data_type: str = 'DICT') -> list:
        """
        [ 执行SQLite查询句柄 ]

        """
        result = {
            G_KEYWORD.Common.TransactionFieldKey.Totai: 0,
            G_KEYWORD.Common.TransactionFieldKey.Items: [],
        }
        sql_sentence = str(sql_handle)
        return_field = sql_handle._data_alias

        # 数据库查询
        db_result: list
        if with_lock:
            with self._lock:
                db_result = self._query(sql_sentence)
        else:
            db_result = self._query(sql_sentence)

        # 处理返回数据
        if return_data_type == 'DICT':
            for item in db_result:
                item = dict(zip(return_field, item))
                result[G_KEYWORD.Common.TransactionFieldKey.Totai] += 1
                result[G_KEYWORD.Common.TransactionFieldKey.Items].append(item)
        else:
            for item in db_result:
                result[G_KEYWORD.Common.TransactionFieldKey.Totai] += 1
                result[G_KEYWORD.Common.TransactionFieldKey.Items].append(item)
    
        return result


    # ----------------------------------------------------------------------
    def _execute_sqlite_handle_write(self, sql_handle: Any, with_lock: bool = False) -> int:
        """
        [ 执行SQLite写入句柄 ]

        """
        result: int
        sql_sentence = str(sql_handle)

        # 数据库写入
        if with_lock:
            with self._lock:
                result = self._write(sql_sentence)
        else:
            result = self._write(sql_sentence)

        return result


    # ----------------------------------------------------------------------
    def _query(self, sql_sentence: str) -> list:
        """
        [ 无锁读取 ]

        ---
        描述
        - 执行查询语句, 该函数不会持有数据库锁。

        ---
        参数
        - sql_sentence : { str } - SQLite 查询语句。

        ---
        返回
        - list : 查询结果列表。

        """
        try:
            cur = self._conn.cursor()  # 创建游标
            cur.execute(sql_sentence)  # 执行语句
            result = cur.fetchall()    # 获取结果
            cur.close()                # 注销游标
            return result              # 返回数据

        except Exception as error:
            raise AwakenDataBaseError(str(error))


    # ----------------------------------------------------------------------
    def _write(self, sql: str) -> int:
        """
        [ 无锁写入 ]

        ---
        描述
        - 执行写入语句, 该函数不会持有数据库锁。

        ---
        参数
        - sql : { str } - SQLite 查询语句。

        ---
        返回
        - int : 写入数据后返回该条数据的自增ID。

        """
        try:
            cur = self._conn.cursor()  # 创建数据库游标
            cur.execute(sql)           # 执行查询语句
            add_id = cur.lastrowid     # 获取数据自增ID
            cur.close()                # 注销数据库游标
            return add_id              # 返回数据自增ID

        except Exception as error:
            raise AwakenDataBaseError(str(error))


    # ----------------------------------------------------------------------
    def connect(self):
        """
        [ 数据库连接 ]

        """
        # 创建数据库连接对象
        self._conn = sqlite3.connect(G_CONST.Path.FilePath.EngineeringDatabase, check_same_thread=False)
        # 创建数据库游标
        Cur = self._conn.cursor()
        
        # 构建数据库表
        for key, value in CreateTableSql.__dict__.items():
            if '__' in key:
                continue
            Cur.execute(value)

        # 注销数据库游标
        Cur.close()


    # ----------------------------------------------------------------------
    def commit(self):
        """
        [ 事务保存 ]

        """
        self._conn.commit()


    # ----------------------------------------------------------------------
    def rollback(self):
        """
        [ 事务回滚 ]

        """
        self._conn.rollback()


    # ----------------------------------------------------------------------
    def quit(self):
        """
        [ 关闭数据库连接 ]

        """
        self._conn.close()


# --------------------------------------------------------------------------
class CreateTableSql(object):
    """
    [ 建表语句 ]

    """


    # ----------------------------------------------------------------------
    User = G_SQLITE_STATEMENT_HANDLE.create_table(
        table_name=G_KEYWORD.DataBase.TableName.User,
        table_field={
            G_KEYWORD.DataBase.UserFieldKey.UserId     : G_KEYWORD.Sqlite.DataType.IdKey,
            G_KEYWORD.DataBase.UserFieldKey.UserNumber : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.UserFieldKey.Password   : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.UserFieldKey.UserName   : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.UserFieldKey.Portrait   : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.UserFieldKey.RoleId     : G_KEYWORD.Sqlite.DataType.Int,
            G_KEYWORD.DataBase.UserFieldKey.UserToken  : G_KEYWORD.Sqlite.DataType.Text,
        }
    )
    """
    [ 建表语句 - 用户表 ]

    """


    # ----------------------------------------------------------------------
    Project = G_SQLITE_STATEMENT_HANDLE.create_table(
        table_name=G_KEYWORD.DataBase.TableName.Project,
        table_field={
            G_KEYWORD.DataBase.ProjectFieldKey.Pid         : G_KEYWORD.Sqlite.DataType.IdKey,
            G_KEYWORD.DataBase.ProjectFieldKey.Type        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.ProjectFieldKey.Name        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.ProjectFieldKey.Docs        : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.ProjectFieldKey.State       : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.ProjectFieldKey.CreatedDate : G_KEYWORD.Sqlite.DataType.DateNotNull,
        }
    )
    """
    [ 建表语句 - 项目表 ]

    """


    # ----------------------------------------------------------------------
    Tasks = G_SQLITE_STATEMENT_HANDLE.create_table(
        table_name=G_KEYWORD.DataBase.TableName.Tasks,
        table_field={
            G_KEYWORD.DataBase.TasksFieldKey.Tid         : G_KEYWORD.Sqlite.DataType.IdKey,
            G_KEYWORD.DataBase.TasksFieldKey.Pid         : G_KEYWORD.Sqlite.DataType.IntNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.Type        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.Name        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.Docs        : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.TasksFieldKey.State       : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.CreatedDate : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.ScriptJson  : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.TasksFieldKey.RunNumber   : G_KEYWORD.Sqlite.DataType.IntNotNull,
            G_KEYWORD.DataBase.TasksFieldKey.LastRunDate : G_KEYWORD.Sqlite.DataType.Date,
            G_KEYWORD.DataBase.TasksFieldKey.LastRunTime : G_KEYWORD.Sqlite.DataType.Date,
            G_KEYWORD.DataBase.TasksFieldKey.SpendTime   : G_KEYWORD.Sqlite.DataType.Int,
        }
    )
    """
    [ 建表语句 - 任务表 ]

    """


    # ----------------------------------------------------------------------
    TaskRecord = G_SQLITE_STATEMENT_HANDLE.create_table(
        table_name=G_KEYWORD.DataBase.TableName.TaskRecord,
        table_field={
            G_KEYWORD.DataBase.TaskRecordFieldKey.TRid        : G_KEYWORD.Sqlite.DataType.IdKey,
            G_KEYWORD.DataBase.TaskRecordFieldKey.Tid         : G_KEYWORD.Sqlite.DataType.IntNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.Name        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.Docs        : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.TaskRecordFieldKey.Type        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.CreatedDate : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.State       : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.TaskRecordFieldKey.StartTime   : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.EndTime     : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.TaskRecordFieldKey.SpendTime   : G_KEYWORD.Sqlite.DataType.Int,
            G_KEYWORD.DataBase.TaskRecordFieldKey.Process     : G_KEYWORD.Sqlite.DataType.Text,
        }
    )
    """
    [ 建表语句 - 任务记录表 ]

    """


    # ----------------------------------------------------------------------
    CaseRecord = G_SQLITE_STATEMENT_HANDLE.create_table(
        table_name=G_KEYWORD.DataBase.TableName.CaseRecord,
        table_field={
            G_KEYWORD.DataBase.CaseRecordFieldKey.CRid        : G_KEYWORD.Sqlite.DataType.IdKey,
            G_KEYWORD.DataBase.CaseRecordFieldKey.TRid        : G_KEYWORD.Sqlite.DataType.IntNotNull,
            G_KEYWORD.DataBase.CaseRecordFieldKey.Name        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.CaseRecordFieldKey.Docs        : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.CaseRecordFieldKey.Type        : G_KEYWORD.Sqlite.DataType.TextNotNull,
            G_KEYWORD.DataBase.CaseRecordFieldKey.CreatedDate : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.CaseRecordFieldKey.State       : G_KEYWORD.Sqlite.DataType.Text,
            G_KEYWORD.DataBase.CaseRecordFieldKey.StartTime   : G_KEYWORD.Sqlite.DataType.DateNotNull,
            G_KEYWORD.DataBase.CaseRecordFieldKey.EndTime     : G_KEYWORD.Sqlite.DataType.Date,
            G_KEYWORD.DataBase.CaseRecordFieldKey.SpendTime   : G_KEYWORD.Sqlite.DataType.Int,
            G_KEYWORD.DataBase.CaseRecordFieldKey.Process     : G_KEYWORD.Sqlite.DataType.Text,
        }
    )
    """
    [ 建表语句 - 用例记录表 ]

    """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 转换格式相关函数 ]

"""
import re
import json

from ..error.type_error import AwakenTypeDetectionFailedError


# --------------------------------------------------------------------------
def try_converting_string_format(string: str):
    """ 
    [ 尝试将字符串转换为对应格式 ]

    ---
    参数
    - string : { str } - 参与计算的字符串。

    """
    # 如果值是字典, 则使用解析成json串返回
    if isinstance(string, dict):
        string = json.dumps(string, ensure_ascii=False).replace("'", r'\"')
        return string

    # 如果参与计算的值不是字符串则尝试转换成字符串
    if not isinstance(string, str):
        try:
            string = str(string)
        except BaseException:
            AwakenTypeDetectionFailedError(expect='str', result=type(string))

    # 字符串规则
    # 如果字符串被 <'> 或 <"> 符号包裹, 则视其为字符串类型
    if string[0] == "'" and string[-1] == "'":
        string_extract = re.findall(r'\'(.*)\'', string)
        return string_extract[0] if len(string_extract) > 0 else ''

    if string[0] == '"' and string[-1] == '"':
        string_extract = re.findall(r'\"(.*)\"', string)
        return string_extract[0] if len(string_extract) > 0 else ''

    # 布尔值规则
    if string in ['true', 'True', 'TRUE']:
        return True
    if string in ['false', 'False', 'FALSE']:
        return False

    # 整数与浮点数规则
    try:
        int_extract = float(string)
        if int_extract.is_integer():
            int_extract = int(int_extract)
        return int_extract

    except ValueError:
        return string



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SQLite 语句关键字
    
"""


# --------------------------------------------------------------------------
class Event:
    """
    [ SQLite 事件关键字 ]

    """
    CreateTable = 'CREATE TABLE'
    """ 建表事件 """


# --------------------------------------------------------------------------
class SelectEvent:
    """
    [ SQLite 查表事件相关 ]

    """
    Count = 'COUNT'
    """ 计算数量 """

    Sum   = 'SUM'
    """ 计算全额 """


# --------------------------------------------------------------------------
class SelectEvent:
    """
    [ SQLite 查表事件相关 ]

    """
    Count = 'COUNT'
    """ 计算数量 """

    Sum   = 'SUM'
    """ 计算全额 """

    Distinct = 'DISTINCT'
    """ 去重 """


# --------------------------------------------------------------------------
class SortType:
    """
    [ SQLite 排序类型 ]

    """
    Asc = 'ASC'
    """ 从小到大, 升序排列 """

    Desc = 'DESC'
    """ 从大到小, 降序排列 """


# --------------------------------------------------------------------------
class CreateTable:
    """
    [ SQLite 建表事件相关 ]

    """
    IfNotExists = 'IF NOT EXISTS'
    """ 建表时如果不存在则创建 """


# --------------------------------------------------------------------------
class DataType:
    """
    [ SQLite 数据类型相关 ]

    """
    IdKey = 'INTEGER PRIMARY KEY'
    """ 整数主键 """

    Int = 'INT'
    IntNotNull = 'INT NOT NULL'

    Text = 'TEXT'
    TextNotNull = 'TEXT NOT NULL'

    Date = 'DATE'
    DateNotNull = 'DATE NOT NULL'



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ Awaken脚本语言解释器 ]

"""
from pathlib import Path

from .base_code_converter import BaseCodeConverter
from ...baseic.keyword import G_KEYWORD
from ...baseic.common.configure import G_CONFIGURE
from ...baseic.structural.interpreter import AwakenTask
from ...baseic.error.interpreter import AwakenInterpreterFileIsEmpty


class GrammarParser: ...


# --------------------------------------------------------------------------
class GrammarParser:
    """
    [ Awaken脚本语言解释器 ]

    """
    _base_code_converter: BaseCodeConverter
    """ 底层编码转换类 """

    
    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        self._base_code_converter = BaseCodeConverter()


    # ----------------------------------------------------------------------
    def parsing_file(self, script_file_path: str | Path):
        """
        [ 解析自文件 ]

        ---
        描述
        - 解析本地的 Awaken 文件, 值得注意的是, 运行本地任务不会写入到数据库。

        ---
        参数
        - script_file_path : { str | Path } - 脚本文件路径。

        ---
        返回
        - AwakenTask : Awaken任务对象。

        """
        translation = self._base_code_converter.convert_file(script_file_path)

        if len(translation.codelines) == 0:
            raise AwakenInterpreterFileIsEmpty

        if G_CONFIGURE.get(G_KEYWORD.Common.Config.Debug):
            print('-------')
            for code in translation.codelines:
                print(code)
            print('-------')

        return AwakenTask(translation, False, 0)


    # ----------------------------------------------------------------------
    def parsing_json(self, script_codelines: list, task_type: str, database_task_id: int):
        """
        [ 解析自语句集 ]

        ---
        描述
        - 解析本地的 Awaken 文件, 值得注意的是, 运行本地任务不会写入到数据库。

        ---
        描述:
            NULL

        """
        translation = self._base_code_converter.convert_codelines(script_codelines, task_type)

        if len(translation.codelines) == 0:
            raise AwakenInterpreterFileIsEmpty

        if G_CONFIGURE.get(G_KEYWORD.Common.Config.Debug):
            print('-------')
            for code in translation.codelines:
                print(code)
            print('-------')

        return AwakenTask(translation, True, database_task_id)



# --------------------------------------------------------------------------
GRAMMAR_PARSER = GrammarParser()
""" Awaken脚本语言解释器实例 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
池运行时

"""
from multiprocessing.pool import ThreadPool
from multiprocessing.dummy import Pool
from multiprocessing.dummy import Lock
from multiprocessing.dummy import Queue
from multiprocessing.dummy import Manager

from ...baseic.keyword import G_KEYWORD
from ...baseic.common.configure import G_CONFIGURE
from ...baseic.decorator.baseic import deco_singleton


# --------------------------------------------------------------------------
@deco_singleton
class PoolRuntime(object):
    """
    [ POOL 运行时 ]

    """
    _pool: ThreadPool
    """ 进程池 """
    _state: bool
    """ 运行状态 """
    _engine_max_count: int
    """ 引擎最大数量 """
    _task_max_count: int
    """ 任务最大数量 """
    _sign_engine_count: int
    """ 注册的驱动数量 """
    _engine_queue: Queue
    """ 引擎队列 """
    _task_queue: Queue
    """ 任务队列 """
    _state_lock: Lock
    """ 运行状态互斥锁 """
    _engine_queue_lock: Lock
    """ 引擎队列互斥锁 """
    _task_queue_lock: Lock
    """ 任务队列互斥锁 """


    # ----------------------------------------------------------------------
    def __init__(self, 
        engine_max_count: int = G_CONFIGURE.get(G_KEYWORD.Common.Config.EngineQueueMaxCount), 
        task_max_count: int = G_CONFIGURE.get(G_KEYWORD.Common.Config.TaskQueueMaxCount)
    ) -> None:
        """
        [ POOL 运行时 ]

        ---
        参数
        - engine_max_count : { int } - 引擎最大数量。
        - task_max_count   : { int } - 任务最大数量。

        """
        self._engine_max_count = engine_max_count
        self._task_max_count = task_max_count

        manager: Manager = Manager()
        self._pool = Pool(self._engine_max_count + 1)
        self._state = False
        self._sign_engine_count = 0
        self._engine_queue = manager.Queue(maxsize=self._engine_max_count)
        self._task_queue = manager.Queue(maxsize=self._task_max_count)
        self._state_lock = manager.Lock()
        self._engine_queue_lock = manager.Lock()
        self._task_queue_lock = manager.Lock()


    # ----------------------------------------------------------------------
    @property
    def pool(self) -> ThreadPool:
        """
        [ 进程池 ]

        """
        return self._pool


    # ----------------------------------------------------------------------
    @property
    def state(self) -> bool:
        """
        [ 运行状态 ]

        """
        with self._state_lock:
            return self._state


    # ----------------------------------------------------------------------
    @property
    def engine_max_count(self) -> int:
        """
        [ 驱动队列最大值 ]

        """
        return self._engine_max_count


    # ----------------------------------------------------------------------
    @property
    def task_max_count(self) -> int:
        """
        [ 任务队列最大值 ]

        """
        return self._task_max_count


    # ----------------------------------------------------------------------
    @property
    def sign_driver_count(self) -> int:
        """
        [ 注册的核心数量 ]

        """
        return self._sign_engine_count


    # ----------------------------------------------------------------------
    @property
    def engine_queue(self) -> Queue:
        """
        [ 核心队列 ]

        """
        with self._engine_queue_lock:
            return self._engine_queue


    # ----------------------------------------------------------------------
    @property
    def task_queue(self) -> Queue:
        """
        [ 任务队列 ]

        """
        with self._task_queue_lock:
            return self._task_queue


    # ----------------------------------------------------------------------
    def set_state(self, state: bool):
        """
        [ 设置运行状态 ]
        
        ---
        参数:
            state { bool } : 运行状态

        """
        with self._state_lock:
            self._state = state


    # ----------------------------------------------------------------------
    def sign_engine(self):
        """
        [ 注册引擎 ]

        """
        self._sign_engine_count += 1


# --------------------------------------------------------------------------
POOL_RUNTIME: PoolRuntime = PoolRuntime()
""" POOL运行时全局实例 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
命令行指令 - 创建项目

"""
from ..baseic.common.template_rendering import G_TEMPLATE_RENDERING
from ..baseic.environment.project_file_handle import G_PROJECT_FILE_HANDLE


# --------------------------------------------------------------------------
def instruction_create_project(argv: list):
    """
    [ 命令行指令 - 创建项目 ]

    ---
    参数
    - argv : { list } - 参数列表。

    """
    if len(argv) >= 1:
        projects = argv[0:]
        
        G_PROJECT_FILE_HANDLE.project_create(projects)
        G_TEMPLATE_RENDERING.render_print(
            title='Awaken - 创建项目',
            is_show_number=False,
            source=[
                '项目创建成功:',
                *projects
            ],
        )

    else:
        G_TEMPLATE_RENDERING.render_print(
            title='Awaken - 创建项目',
            is_show_number=False,
            source=[
                '创建项目指令参数异常 !',
                '示例:',
                '>> awaken -make project',
                '>> awaken -make project1 project2',
            ],
        )
    exit(0)



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
服务执行池

"""
import gc
import psutil

from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer

from .pool_runtime import POOL_RUNTIME
from ...baseic.keyword import G_KEYWORD
from ...baseic.common.configure import G_CONFIGURE
from ...baseic.broadcast.awaken_log import G_LOG
from ...server.api.awaken_app import AwakenApp
from ...kit.engine_agency import EngineAgency
from ...baseic.structural.interpreter import AwakenTask


# --------------------------------------------------------------------------
class PerformServer(object):
    """
    [ 执行服务 ]

    ---
    描述
    - 开启时, 将在本地挂起一个执行池进程与 WebAPI 服务。

    """
    _debug: bool
    """ 调试模式 """


    # ----------------------------------------------------------------------
    def running(self, 
        port: int = 3700, 
        debug: bool = G_CONFIGURE.get(G_KEYWORD.Common.Config.Debug)
    ):
        """
        [ 运行 ]

        ---
        参数:
        - port  : { int }  - 服务端口号, 默认为 3700。
        - debug : { bool } - 调试模式。

        """
        self._debug = debug
        POOL_RUNTIME.pool.apply_async(func=self.monitor)
        
        app = AwakenApp()
        http_server = HTTPServer(WSGIContainer(app.living))
        http_server.listen(port)

        if self._debug:
            G_LOG.info(f'执行服务运行中 >> 端口号 :: [ {str(port)} ] !')

        IOLoop.current().start()


    # ----------------------------------------------------------------------
    def monitor(self):
        """
        [ 监听进程 ]
        
        ---
        描述
        - 监听来自 POOL 运行时的队列信息。

        """
        if self._debug:
            G_LOG.info('监听进程已启动 ...')

        while True:
            # --------------------------------------------------------------
            # 调用多进程执行任务
            # 驱动冷启动时会占用大量的 CPU 资源, 为避免因为 CPU 占用率过高导致进程启动异常
            # 每隔 X 秒检查 CPU 占用率, 只有 CPU 占用率少于 X 时才会启动进程
            # --------------------------------------------------------------
            task = POOL_RUNTIME.task_queue.get()
            engine = POOL_RUNTIME.engine_queue.get()
            while True:
                if psutil.cpu_percent(G_CONFIGURE.get(G_KEYWORD.Common.Config.CpuPropertyTime)) <= G_CONFIGURE.get(G_KEYWORD.Common.Config.CpuPropertyCeiling):
                    POOL_RUNTIME.pool.apply_async(func=self._emit, args=[task, engine])
                    break


    # ----------------------------------------------------------------------
    def _emit(self, task: AwakenTask, engine: EngineAgency):
        """
        [ 触发工作流 ]
        
        ---
        参数:
            task   : { AwakenTask }   - 任务对象。
            engine : { EngineAgency } - 驱动对象。

        """
        # 核心执行任务
        engine.start(task)
        # 核心执行完任务后, 放回核心队列等待
        POOL_RUNTIME.engine_queue.put(engine)
        # 强制 GC 回收垃圾避免内存溢出
        gc.collect()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 任务相关接口 ]

"""
import time
import json

from flask import Response
from flask import Blueprint

from ..server_request_handler import URL_PREFIX
from ..server_request_handler import ServerRequestHandler
from ...db import DB
from ...db import SQL
from ....baseic.const import G_CONST
from ....baseic.keyword import G_KEYWORD
from ....baseic.decorator.web_api import deco_webapi_error_capture


blueprint_tasks = Blueprint('tasks', __name__)
""" 任务接口蓝图 """


POST_TASKS_GET_TASKS_COUNT     = '/tasks/count'
""" 接口URL - 获取任务数量 """ 
POST_TASKS_GET_TASKS_LIST      = '/tasks/list'
""" 接口URL - 获取任务列表 """
POST_TASKS_GET_TASKS_CREATION  = '/tasks/creation'
""" 接口URL - 任务创建 """ 
POST_TASKS_GET_TASKS_MODIFY    = '/tasks/modify'
""" 接口URL - 任务编辑 """ 
POST_TASKS_GET_TASKS_DELETE    = '/tasks/delete'
""" 接口URL - 任务删除 """


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_tasks.route(''.join([URL_PREFIX, POST_TASKS_GET_TASKS_COUNT]), methods=['POST'])
def post_count() -> Response:
    """
    [ POST - 查询任务数量 ]

    ---
    可选参数
    - tid           : { str } - 任务编号。
    - pid           : { str } - 项目编号。
    - type          : { str } - 任务类型。
    - name          : { str } - 任务名称。
    - state         : { str } - 任务状态。
    - created_date  : { str } - 任务创建日期, 例如 2022-02-02。
    - run_number    : { str } - 运行次数。
    - last_run_date : { str } - 最后运行日期。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        optional_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid,
            G_KEYWORD.Api.TasksFieldKey.Pid,
            G_KEYWORD.Api.TasksFieldKey.Type,
            G_KEYWORD.Api.TasksFieldKey.Name,
            G_KEYWORD.Api.TasksFieldKey.State,
            G_KEYWORD.Api.TasksFieldKey.CreatedDate,
            G_KEYWORD.Api.TasksFieldKey.RunNumber,
            G_KEYWORD.Api.TasksFieldKey.LastRunDate,
        ]
    )

    # 从数据库中获取数据
    db_data = DB.execute(SQL.Tasks.get_count(paras), return_data_type='LIST')
    count = db_data.get(G_KEYWORD.Common.TransactionFieldKey.Items)[0][0]

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Count: count
    })
    return ServerRequestHandler.successful(result)


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_tasks.route(''.join([URL_PREFIX, POST_TASKS_GET_TASKS_LIST]), methods=['POST'])
def post_list() -> Response:
    """
    [ POST - 查询任务列表 ]

    ---
    可选参数
    - tid           : { str } - 任务编号。
    - pid           : { str } - 项目编号。
    - type          : { str } - 任务类型。
    - name          : { str } - 任务名称。
    - state         : { str } - 任务状态。
    - created_date  : { str } - 任务创建日期, 例如 2022-02-02。
    - run_number    : { str } - 运行次数。
    - last_run_date : { str } - 最后运行日期。
    - page          : { str } - 当前页数。
    - number        : { str } - 单页数据量。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        optional_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid,
            G_KEYWORD.Api.TasksFieldKey.Pid,
            G_KEYWORD.Api.TasksFieldKey.Type,
            G_KEYWORD.Api.TasksFieldKey.Name,
            G_KEYWORD.Api.TasksFieldKey.State,
            G_KEYWORD.Api.TasksFieldKey.CreatedDate,
            G_KEYWORD.Api.TasksFieldKey.RunNumber,
            G_KEYWORD.Api.TasksFieldKey.LastRunDate,
            G_KEYWORD.Common.PagingDataFieldKey.Page,
            G_KEYWORD.Common.PagingDataFieldKey.Number,
        ]
    )

    # 从数据库中获取数据
    db_data = DB.execute(SQL.Tasks.get_list(paras))
    items = db_data.get(G_KEYWORD.Common.TransactionFieldKey.Items)
    totai = db_data.get(G_KEYWORD.Common.TransactionFieldKey.Totai)

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Items: items,
        G_KEYWORD.Api.TransactionFieldKey.Totai: totai,
    })
    return ServerRequestHandler.successful(result)


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_tasks.route(''.join([URL_PREFIX, POST_TASKS_GET_TASKS_CREATION]), methods=['POST'])
def post_creation() -> Response:
    """
    [ POST - 任务创建 ]

    ---
    必要参数
    - pid  : { int } - 项目编号。
    - type : { int } - 任务类型。
    - name : { int } - 任务名称。

    ---
    可选参数:
    - docs : { str } - 任务说明。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Pid,
            G_KEYWORD.Api.TasksFieldKey.Type,
            G_KEYWORD.Api.TasksFieldKey.Name,
        ],
        optional_keys=[
            G_KEYWORD.Api.TasksFieldKey.Docs,
        ]
    )
    if not paras.get(G_KEYWORD.Api.TasksFieldKey.Docs):
        paras.update({G_KEYWORD.Api.TasksFieldKey.Docs : 'NULL'})

    # 从数据库中写入数据
    tid = DB.execute(SQL.Tasks.creation(
        pid =paras.get(G_KEYWORD.Api.TasksFieldKey.Pid),
        type=paras.get(G_KEYWORD.Api.TasksFieldKey.Type),
        name=paras.get(G_KEYWORD.Api.TasksFieldKey.Name),
        docs=paras.get(G_KEYWORD.Api.TasksFieldKey.Docs),
        created_date=time.strftime('%Y-%m-%d'),
        state=G_CONST.State.Project.Null,  # 状态, 暂时不实现
    ))
    DB.commit()

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Message: '任务创建成功 !',
        G_KEYWORD.Api.TasksFieldKey.Tid: tid
    })
    return ServerRequestHandler.successful(result)


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_tasks.route(''.join([URL_PREFIX, POST_TASKS_GET_TASKS_MODIFY]), methods=['POST'])
def post_modify() -> Response:
    """
    [ POST - 任务编辑 ]

    ---
    必要参数
    - tid         : { int } - 任务编号。
    - name        : { int } - 任务名称。
    - docs        : { str } - 任务说明。
    - script_json : { Any } - 脚本JSON对象。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid,
            G_KEYWORD.Api.TasksFieldKey.Name,
            G_KEYWORD.Api.TasksFieldKey.Docs,
            G_KEYWORD.Api.TasksFieldKey.ScriptJson,
        ]
    )

    script_json = paras.get(G_KEYWORD.Api.TasksFieldKey.ScriptJson)
    script_json_string = json.dumps(script_json, ensure_ascii=False)

    DB.execute(SQL.Tasks.modify(
        tid =paras.get(G_KEYWORD.Api.TasksFieldKey.Tid), 
        name=paras.get(G_KEYWORD.Api.TasksFieldKey.Name), 
        docs=paras.get(G_KEYWORD.Api.TasksFieldKey.Docs), 
        script_json_string=script_json_string
    ))
    DB.commit()

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Message: '任务编辑成功 !'
    })
    return ServerRequestHandler.successful(result)


# --------------------------------------------------------------------------
@deco_webapi_error_capture
@blueprint_tasks.route(''.join([URL_PREFIX, POST_TASKS_GET_TASKS_DELETE]), methods=['POST'])
def post_delete() -> Response:
    """
    [ POST - 任务删除 ]

    ---
    必要参数
    - tid : { int } - 任务编号。

    """
    result = {}

    # 解析请求
    paras = ServerRequestHandler.analysis_request_parameter(
        must_keys=[
            G_KEYWORD.Api.TasksFieldKey.Tid
        ]
    )
    tid = paras.get(G_KEYWORD.Api.TasksFieldKey.Tid)

    # 从数据库中写入数据
    DB.execute(SQL.Tasks.delete(tid=tid))
    DB.commit()

    # 处理返回
    result.update({
        G_KEYWORD.Api.TransactionFieldKey.Message: '任务删除成功 !'
    })
    return ServerRequestHandler.successful(result)



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ 静态执行池 ]

"""
import gc
import psutil

from .pool_runtime import PoolRuntime
from ...kit.engine_agency import EngineAgency
from ...baseic.keyword import G_KEYWORD
from ...baseic.common.configure import G_CONFIGURE
from ...baseic.error.baseic import G_ERROR_RECORDER
from ...baseic.broadcast.awaken_log import G_LOG
from ...baseic.decorator.baseic import deco_singleton
from ...baseic.structural.interpreter import AwakenTask


# --------------------------------------------------------------------------
@deco_singleton
class PerFormPool(object):
    """
    [ 静态执行池 ]

    ---
    描述
    - 通常用于静态执行多进程任务。

    """
    _debug: bool
    """ 调试模式 """
    _pool_runtime: PoolRuntime
    """ 运行时 """


    # ----------------------------------------------------------------------
    def __init__(self, 
        engine_max_count: int = 3,
        task_max_count: int = 50,
        debug: bool = G_CONFIGURE.get(G_KEYWORD.Common.Config.Debug)
    ) -> None:
        """
        [ 静态执行池 ]

        ---
        参数
        - engine_max_count : { int }  - 引擎最大数量。
        - task_max_count   : { int }  - 任务最大数量。
        - debug            : { bool } - 调试模式。

        """
        self._debug = debug
        self._pool_runtime = PoolRuntime(engine_max_count, task_max_count)

        # 创建驱动代理
        for _ in range(engine_max_count):
            self._pool_runtime.engine_queue.put(EngineAgency())
            self._pool_runtime.sign_engine()


    # ----------------------------------------------------------------------
    def put_task(self, task: AwakenTask):
        """
        [ 推送到任务队列 ]
        
        ---
        参数
        - task : { AwakenTask } - 任务对象。

        """
        count = self._pool_runtime.task_queue.qsize()
        max_count = self._pool_runtime.task_queue.maxsize
        if max_count > count:
            self._pool_runtime.task_queue.put(task)


    # ----------------------------------------------------------------------
    def running(self):
        """
        [ 运行 ]

        """
        # ------------------------------------------------------------------
        # 开始一个监听循环
        # 当运行时的任务队列为空且所有代理引擎都空闲时, 跳出监听循环
        # ------------------------------------------------------------------
        while True:

            if self._pool_runtime.task_queue.qsize() == 0:
                if self._debug:
                    G_LOG.info('任务队列消耗完毕, 正在等待任务执行 !')

                while True:
                    core_count = self._pool_runtime.engine_queue.qsize()
                    if core_count == self._pool_runtime.sign_driver_count:
                        if self._debug:
                            G_LOG.info('任务全部执行完毕, 退出程序 !')
                        break
                
                if G_ERROR_RECORDER.get_error_number() > 0:
                    G_ERROR_RECORDER.template_out()

                break

            # --------------------------------------------------------------
            # 调用多进程执行任务
            # 驱动冷启动时会占用大量的 CPU 资源, 为避免因为 CPU 占用率过高导致进程启动异常
            # 每隔 X 秒检查 CPU 占用率, 只有 CPU 占用率少于 X 时才会启动进程
            # --------------------------------------------------------------
            task = self._pool_runtime.task_queue.get()
            engine = self._pool_runtime.engine_queue.get()
            while True:
                if psutil.cpu_percent(G_CONFIGURE.get(G_KEYWORD.Common.Config.CpuPropertyTime)) <= G_CONFIGURE.get(G_KEYWORD.Common.Config.CpuPropertyCeiling):
                    self._pool_runtime.pool.apply_async(func=self._emit, args=[task, engine])
                    break


    # ----------------------------------------------------------------------
    def _emit(self, task: AwakenTask, engine: EngineAgency):
        """
        [ 触发工作流 ]

        ---
        参数
        - task   : { AwakenTask }   - 任务对象。
        - engine : { EngineAgency } - 驱动对象。

        """
        # 核心执行任务
        engine.start(task)
        # 核心执行完任务后, 放回核心队列等待
        self._pool_runtime.engine_queue.put(engine)
        # 强制 GC 回收垃圾避免内存溢出
        gc.collect()



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
脚本字段关键字
    
"""


# --------------------------------------------------------------------------
class Namespace:
    """
    [ 命名空间关键字 ]

    ---
    描述
    - 脚本的命名空间相关的字段。

    """
    TaskType = 'TaskType'
    """ 任务类型 """

    TaskName = 'TaskName'
    """ 任务名称 """

    TaskDocs = 'TaskDocs'
    """ 任务注释 """


# --------------------------------------------------------------------------
class Decorator:
    """ 
    [ 装饰器关键字 ]

    ---
    描述
    - 脚本的装饰器相关的字段。

    """
    BrowserType = 'BrowserType'
    """ 浏览器类型 """


# --------------------------------------------------------------------------
class ScriptJson:
    """
    [ 脚本JSON对象关键字 ]

    ---
    描述
    - 脚本的JSON对象相关的字段。

    """
    Type = 'type'
    Name = 'name'
    Docs = 'docs'
    
    Namespace = 'namespace'
    """ 命名空间 """

    Cases = 'cases'
    """ 用例集 """



# Copyright 2022 quinn.7@foxmail.com All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
[ SQL语句映射 ]

"""
from .blueprint.sql_user import SqlUser
from .blueprint.sql_project import SqlProject
from .blueprint.sql_tasks import SqlTasks
from .blueprint.sql_record_task import SqlTaskRecord
from .blueprint.sql_record_case import SqlCaseRecord


# --------------------------------------------------------------------------
class AwakenSql(object):
    """
    [ SQL语句映射类 ]

    """


    # ----------------------------------------------------------------------
    @property
    def User(self) -> SqlUser:
        """
        [ 用户相关 ]

        """
        return SqlUser


    # ----------------------------------------------------------------------
    @property
    def Project(self) -> SqlProject:
        """
        [ 项目相关 ]

        """
        return SqlProject


    # ----------------------------------------------------------------------
    @property
    def Tasks(self) -> SqlTasks:
        """
        [ 任务相关 ]

        """
        return SqlTasks


    # ----------------------------------------------------------------------
    @property
    def TaskRecord(self) -> SqlTaskRecord:
        """
        [ 任务记录相关 ]

        """
        return SqlTaskRecord


    # ----------------------------------------------------------------------
    @property
    def CaseRecord(self) -> SqlCaseRecord:
        """
        [ 用例记录相关 ]

        """
        return SqlCaseRecord
